from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.cruds._base_cruds import T_ID


def _print_ids_or_length(resource_ids: SequenceNotStr[T_ID], limit: int = 10) -> str:
    if len(resource_ids) == 1:
        return f"{resource_ids[0]!r}"
    elif len(resource_ids) <= limit:
        return f"{resource_ids}"
    else:
        return f"{len(resource_ids)} items"



import difflib
from collections.abc import Collection, Iterable, Iterator, Sequence, Set
from itertools import islice
from typing import Any, TypeVar

from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump


def flatten_dict(dct: dict[str, Any]) -> dict[tuple[str, ...], Any]:
    """Flatten a dictionary to a list of tuples with the key path and value."""
    items: dict[tuple[str, ...], Any] = {}
    for key, value in dct.items():
        if isinstance(value, dict):
            for sub_key, sub_value in flatten_dict(value).items():
                items[(key, *sub_key)] = sub_value
        else:
            items[(key,)] = value
    return items


def flatten_dict_json_path(dct: dict[str, Any], keep_structured: Set[str] | None = None) -> dict[str, Any]:
    """Flatten a dictionary to a dictionary with JSON path keys.

    Empty keys are ignored.

    Args:
        dct: The dictionary to flatten.
        keep_structured: A set of keys to keep structured (not flatten). If a key is in this set,
            it will not be flattened, and the value will be kept as is. This only applies to top-level keys.

    Returns:
        A dictionary with JSON path keys.
    """

    return _flatten(dct, keep_structured=keep_structured or set())


def _flatten(obj: Any, keep_structured: Set[str], path: str = "") -> dict[str, Any]:
    items: dict[str, Any] = {}
    if isinstance(obj, dict):
        for key, value in obj.items():
            if not key:
                continue
            current_path = f"{path}.{key}" if path else key
            if key in keep_structured:
                items[current_path] = value
            else:
                items.update(_flatten(value, set(), current_path))
    elif isinstance(obj, list):
        for i, value in enumerate(obj):
            current_path = f"{path}[{i}]"
            if isinstance(value, (dict, list)):
                items.update(_flatten(value, set(), current_path))
            else:
                items[current_path] = value
    else:
        items[path] = obj

    return items


def to_diff(a: dict[str, Any], b: dict[str, Any]) -> Iterator[str]:
    a_str = yaml_safe_dump(a, sort_keys=True)
    b_str = yaml_safe_dump(b, sort_keys=True)

    return difflib.unified_diff(a_str.splitlines(), b_str.splitlines())


def in_dict(keys: Iterable[str], dictionary: dict) -> bool:
    return all(key in dictionary for key in keys)


def humanize_collection(collection: Collection[Any], /, *, sort: bool = True, bind_word: str = "and") -> str:
    if not collection:
        return ""
    elif len(collection) == 1:
        return str(next(iter(collection)))

    strings = (str(item) for item in collection)
    if sort:
        sequence = sorted(strings)
    else:
        sequence = list(strings)

    return f"{', '.join(sequence[:-1])} {bind_word} {sequence[-1]}"


def chunker(iterable: Iterable[Any], size: int) -> Iterator[list[Any]]:
    iterator = iter(iterable)
    while chunk := list(islice(iterator, size)):
        yield chunk


T_Sequence = TypeVar("T_Sequence", bound=Sequence)


def chunker_sequence(sequence: T_Sequence, size: int) -> Iterator[T_Sequence]:
    """Yield successive n-sized chunks from sequence."""
    for i in range(0, len(sequence), size):
        # MyPy does not expect sequence[i : i + size] to be of type T_Sequence
        yield sequence[i : i + size]  # type: ignore[misc]



from pydantic import Field

from .base import ToolkitResource


class LabelsYAML(ToolkitResource):
    external_id: str = Field(
        description="The external ID provided by the client.",
        max_length=255,
    )
    name: str = Field(
        description="The name of the label.",
        min_length=1,
        max_length=140,
    )
    description: str | None = Field(
        default=None,
        description="The description of the label.",
        min_length=1,
        max_length=500,
    )
    data_set_external_id: str | None = Field(
        default=None,
        description="The id of the dataset this label belongs to.",
    )



import sys
from abc import ABC
from types import MappingProxyType
from typing import Any, ClassVar, Literal, cast

from pydantic import Field, JsonValue, ModelWrapValidatorHandler, field_validator, model_serializer, model_validator
from pydantic_core.core_schema import SerializationInfo, SerializerFunctionWrapHandler

from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses

from .base import BaseModelResource, ToolkitResource

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


class JobFormat(BaseModelResource, ABC):
    type: ClassVar[str]
    encoding: Literal["utf8", "utf16", "utf16le", "latin1"] = Field(
        "utf8", description="The type of encoding to convert from."
    )
    compression: Literal["gzip"] = Field(
        "gzip",
        description="The compression applied to incoming messages. The messages are decompressed before being passed to transformations. This is usually not relevant for REST, where this is handled automatically, but MQTT, Kafka, and EventHub have no such mechanisms.",
    )

    @model_validator(mode="wrap")
    @classmethod
    def find_format(cls, data: "dict[str, Any] | JobFormat", handler: ModelWrapValidatorHandler[Self]) -> Self:
        if isinstance(data, JobFormat):
            return cast(Self, data)
        if not isinstance(data, dict):
            raise ValueError(f"Invalid input for format '{type(data)}' expected dict")

        if cls is not JobFormat:
            # We are already in a subclass, so just validate as usual
            return handler(data)
        # If not we need to find the right subclass based on the type field.
        if "type" not in data:
            raise ValueError("Invalid input format missing 'type' key")
        type_ = data["type"]
        if type_ not in _JOB_FORMAT_CLS_BY_TYPE:
            raise ValueError(
                f"invalid type '{type_}'. Expected one of {humanize_collection(_JOB_FORMAT_CLS_BY_TYPE.keys(), bind_word='or')}"
            )
        cls_ = _JOB_FORMAT_CLS_BY_TYPE[type_]
        return cast(Self, cls_.model_validate({k: v for k, v in data.items() if k != "type"}))

    @model_serializer(mode="wrap", when_used="always", return_type=dict)
    def include_type(self, handler: SerializerFunctionWrapHandler) -> dict:
        serialized_data = handler(self)
        serialized_data["type"] = self.type
        return serialized_data


class CustomFormat(JobFormat):
    type: ClassVar[str] = "custom"
    mapping_id: str = Field(description="ID of the mapping this format should be tied to.", max_length=255)


class PrefixConfig(BaseModelResource):
    from_topic: bool | None = Field(None, description="Generate the prefix based on the topic of the received message.")
    prefix: str | None = Field(
        None,
        description="A fixed prefix to the generated IDs.",
        max_length=255,
    )


class SpaceRef(BaseModelResource):
    space: str = Field(description="The data models space where time series will be created.")


class DataModelFormat(JobFormat, ABC):
    prefix: PrefixConfig | None = Field(
        None,
        description="Generate a prefix for resources created using this format. If both prefix and fromTopic are set, the generated ID will be on the form [prefix][topic][id].",
    )
    data_models: list[SpaceRef] | None = Field(
        None,
        description="Data models configuration to specify the space for all instances.",
        max_length=10,
    )


class CogniteFormat(DataModelFormat):
    type: ClassVar[str] = "cognite"


class RockwellFormat(DataModelFormat):
    type: ClassVar[str] = "rockwell"


class ValueFormat(DataModelFormat):
    type: ClassVar[str] = "value"


class MQTTConfig(BaseModelResource):
    topic_filter: str = Field(description="Topic filter")


class KafkaConfig(BaseModelResource):
    topic: str = Field(description="Kafka topic to connect to", max_length=200)
    partitions: int = Field(1, description="Number of partitions on the topic.", ge=1, le=10)


class IncrementalLoad(BaseModelResource, ABC):
    type: ClassVar[str]

    @model_validator(mode="wrap")
    @classmethod
    def find_incremental_load(
        cls, data: "dict[str, Any] | IncrementalLoad", handler: ModelWrapValidatorHandler[Self]
    ) -> Self:
        if isinstance(data, IncrementalLoad):
            return cast(Self, data)
        if not isinstance(data, dict):
            raise ValueError(f"Invalid input '{type(data)}'. Expected dict")

        if cls is not IncrementalLoad:
            # We are already in a subclass, so just validate as usual
            return handler(data)
        # If not we need to find the right subclass based on the type field.
        if "type" not in data:
            raise ValueError("Invalid input format missing 'type' key")
        type_ = data["type"]
        if type_ not in _INCREMENTAL_LOAD_CLS_BY_TYPE:
            raise ValueError(
                f"invalid type '{type_}'. Expected one of {humanize_collection(_INCREMENTAL_LOAD_CLS_BY_TYPE.keys(), bind_word='or')}"
            )
        cls_ = _INCREMENTAL_LOAD_CLS_BY_TYPE[type_]
        return cast(Self, cls_.model_validate({k: v for k, v in data.items() if k != "type"}))

    @model_serializer(mode="wrap", when_used="always", return_type=dict)
    def include_type(self, handler: SerializerFunctionWrapHandler) -> dict:
        serialized_data = handler(self)
        serialized_data["type"] = self.type
        return serialized_data


class BodyIncrementalLoad(IncrementalLoad):
    type: ClassVar[str] = "body"
    value: str = Field(
        "Expression yielding next message body. Note that body-based pagination is not allowed to be used if method is not set to post."
    )


class HeaderValueIncrementalLoad(IncrementalLoad):
    type: ClassVar[str] = "headerValue"
    key: str = Field("Key to insert the generated value into")
    value: str = Field("Expression that will be evaluated, and its result used as a header value.")


class NextURLIncrementalLoad(IncrementalLoad):
    type: ClassVar[str] = "nextUrl"
    value: str = Field("Expression yielding the next URL to call.")


class QueryParameterIncrementalLoad(IncrementalLoad):
    type: ClassVar[str] = "queryParameter"
    key: str = Field("Key to insert the generated value into")
    value: str = Field("Expression that will be evaluated, and its result used as a query parameter")


class RestConfig(BaseModelResource):
    interval: Literal["5m", "15m", "1h", "6h", "12h", "1d"]
    path: str = Field(
        description="Path of resource to access on the server, without query.", min_length=1, max_length=2048
    )
    method: Literal["get", "post"] = Field("get", description="HTTP method to use for each request.")
    body: dict[str, JsonValue] | None = Field(
        None,
        description="Initial JSON body to send with request. Only applicable if method is post. Maximum of 10000 bytes total.",
    )
    query: dict[str, str] | None = Field(
        None,
        description="Query parameters to include in request. String key -> String value. Limits: Maximum 255 characters per key, 2048 per value, and at most 32 pairs.",
        max_length=32,
    )
    headers: dict[str, str] | None = Field(
        None,
        description="HTTP headers to include in request. String key -> String value. Limits: Maximum 255 characters per key, 2048 per value, and at most 32 pairs.",
        max_length=32,
    )
    incremental_load: IncrementalLoad | None = Field(
        None,
        description="The format of the messages from the source. This is used to convert messages coming from the source system to a format that can be inserted into CDF.",
    )
    pagination: IncrementalLoad | None = Field(
        None,
        description="The format of the messages from the source. This is used to convert messages coming from the source system to a format that can be inserted into CDF.",
    )

    @field_validator("incremental_load", mode="after")
    def validate_incremental_load_type(cls, v: IncrementalLoad | None) -> IncrementalLoad | None:
        if v is None:
            return v
        allowed_incremental_load_types = {
            BodyIncrementalLoad.type,
            HeaderValueIncrementalLoad.type,
            QueryParameterIncrementalLoad.type,
        }
        if v.type not in allowed_incremental_load_types:
            raise ValueError(
                f"Invalid type '{v.type}'. Expected one of {humanize_collection(allowed_incremental_load_types)}"
            )
        return v

    @model_serializer(mode="wrap")
    def serialize_incremental_load(self, handler: SerializerFunctionWrapHandler, info: SerializationInfo) -> dict:
        # IncrementalLoad and pagination are serialized as empty dict
        # This issue arises because Pydantic's serialization mechanism doesn't automatically
        # handle polymorphic serialization for subclasses of IncrementalLoad.
        # To address this, we include the below to explicitly calling model dump on the input
        serialized_data = handler(self)
        if self.incremental_load is not None:
            serialized_data["incrementalLoad" if info.by_alias else "incremental_load"] = (
                self.incremental_load.model_dump(**vars(info))
            )
        if self.pagination is not None:
            serialized_data["pagination"] = self.pagination.model_dump(**vars(info))
        return serialized_data


class HostedExtractorJobYAML(ToolkitResource):
    external_id: str = Field(
        description="The external ID provided by the client. Must be unique for the resource type.",
        max_length=255,
    )
    destination_id: str = Field(
        description="ID of the destination this job should write to.",
        max_length=255,
    )
    source_id: str = Field(
        description="ID of the source this job should read from.",
        max_length=255,
    )
    format: JobFormat = Field(
        description="The format of the messages from the source. This is used to convert messages coming from the source system to a format that can be inserted into CDF.",
    )
    config: MQTTConfig | KafkaConfig | RestConfig | None = Field(
        None,
        description="Source specific job configuration. The type depends on the type of source, and is required for some sources.",
    )


_INCREMENTAL_LOAD_CLS_BY_TYPE: MappingProxyType[str, type[IncrementalLoad]] = MappingProxyType(
    {load.type: load for load in get_concrete_subclasses(IncrementalLoad)}
)
_JOB_FORMAT_CLS_BY_TYPE: MappingProxyType[str, type[JobFormat]] = MappingProxyType(
    {fmt.type: fmt for fmt in get_concrete_subclasses(JobFormat)}
)



from typing import ClassVar

from cognite.client.data_classes.data_modeling import NodeId

from cognite_toolkit._cdf_tk.client.data_classes.migration import (
    AssetCentricId,
)
from cognite_toolkit._cdf_tk.commands._migrate.issues import (
    ConversionIssue,
    FailedConversion,
    InvalidPropertyDataType,
    ReadAPIIssue,
    ReadFileIssue,
    ReadIssue,
    WriteIssue,
)


class TestMigrationIssues:
    def test_read_file_issue(self) -> None:
        issue = ReadFileIssue(row_no=10, error="Cannot read column 'id' value is not an integer")
        assert issue.dump() == {
            "type": "fileRead",
            "rowNo": 10,
            "error": "Cannot read column 'id' value is not an integer",
        }

    def test_read_api_issue(self) -> None:
        asset_centric_id = AssetCentricId(resource_type="asset", id_=123)
        issue = ReadAPIIssue(asset_centric_id=asset_centric_id, error="API error")
        assert issue.dump() == {
            "type": "apiRead",
            "assetCentricId": {"resourceType": "asset", "id": 123},
            "error": "API error",
        }

    def test_read_issue_subclass(self) -> None:
        class CustomReadIssue(ReadIssue):
            type: ClassVar[str] = "customRead"
            custom_field: str

        issue = CustomReadIssue(custom_field="custom value")
        assert issue.dump() == {
            "type": "customRead",
            "customField": "custom value",
        }

    def test_conversion_issue_minimal(self) -> None:
        asset_centric_id = AssetCentricId(resource_type="asset", id_=456)
        instance_id = NodeId(space="test_space", external_id="test_instance")

        conversion_issue = ConversionIssue(asset_centric_id=asset_centric_id, instance_id=instance_id)

        assert conversion_issue.dump() == {
            "type": "conversion",
            "assetCentricId": {"resourceType": "asset", "id": 456},
            "instanceId": {"space": "test_space", "externalId": "test_instance", "type": "node"},
            "failedConversions": [],
            "invalidInstancePropertyTypes": [],
            "missingAssetCentricProperties": [],
            "missingInstanceProperties": [],
            "ignoredAssetCentricProperties": [],
            "missingInstanceSpace": None,
        }

    def test_conversion_issue_with_all_fields(self) -> None:
        asset_centric_id = AssetCentricId(resource_type="timeseries", id_=789)
        instance_id = NodeId(space="demo_space", external_id="demo_instance")

        failed_conversion = FailedConversion(property_id="value", value="not_a_number", error="Cannot convert to float")
        invalid_property = InvalidPropertyDataType(property_id="status", expected_type="boolean")

        conversion_issue = ConversionIssue(
            asset_centric_id=asset_centric_id,
            instance_id=instance_id,
            missing_asset_centric_properties=["missing_source_prop"],
            missing_instance_properties=["missing_target_prop"],
            invalid_instance_property_types=[invalid_property],
            failed_conversions=[failed_conversion],
        )

        assert conversion_issue.dump() == {
            "type": "conversion",
            "assetCentricId": {"resourceType": "timeseries", "id": 789},
            "instanceId": {"space": "demo_space", "externalId": "demo_instance", "type": "node"},
            "missingAssetCentricProperties": ["missing_source_prop"],
            "missingInstanceProperties": ["missing_target_prop"],
            "invalidInstancePropertyTypes": [{"propertyId": "status", "expectedType": "boolean"}],
            "failedConversions": [{"propertyId": "value", "value": "not_a_number", "error": "Cannot convert to float"}],
            "ignoredAssetCentricProperties": [],
            "missingInstanceSpace": None,
        }

    def test_write_issue_minimal(self) -> None:
        instance_id = NodeId(space="write_space", external_id="write_instance")

        write_issue = WriteIssue(instance_id=instance_id, status_code=400)

        assert write_issue.dump() == {
            "type": "write",
            "instanceId": {"space": "write_space", "externalId": "write_instance", "type": "node"},
            "statusCode": 400,
            "message": None,
        }

    def test_write_issue_with_message(self) -> None:
        instance_id = NodeId(space="error_space", external_id="error_instance")

        write_issue = WriteIssue(instance_id=instance_id, status_code=500, message="Internal server error occurred")

        assert write_issue.dump() == {
            "type": "write",
            "instanceId": {"space": "error_space", "externalId": "error_instance", "type": "node"},
            "statusCode": 500,
            "message": "Internal server error occurred",
        }



from pathlib import Path

from cognite_toolkit._cdf_tk.commands.build_cmd import BuildCommand
from cognite_toolkit._cdf_tk.cruds import HostedExtractorSourceCRUD
from cognite_toolkit._cdf_tk.tk_warnings import MissingRequiredParameterWarning
from tests.data import COMPLETE_ORG


class TestHostedExtractorDestinationLoader:
    def test_hosted_extractor_destination_spec(self, tmp_path: Path) -> None:
        cmd = BuildCommand(print_warning=False)
        cmd.execute(
            verbose=False,
            build_dir=tmp_path,
            organization_dir=COMPLETE_ORG,
            selected=None,
            build_env_name="dev",
            no_clean=False,
        )

        warns = [
            w
            for w in cmd.warning_list
            if isinstance(w, MissingRequiredParameterWarning)
            and HostedExtractorSourceCRUD.folder_name in w.filepath.parts
        ]

        assert len(warns) == 0, warns



import sys

from pydantic import Field, JsonValue, model_validator

from cognite_toolkit._cdf_tk.constants import SPACE_FORMAT_PATTERN

from .base import BaseModelResource, ToolkitResource

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

_SUBSCRIPTION_MAX_TOTAL_TIMESERIES = 10_000


class InstanceId(BaseModelResource):
    space: str = Field(
        description="The Space identifier (id).",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="The external ID of the instance.",
        max_length=255,
    )


class DatapointSubscriptionYAML(ToolkitResource):
    external_id: str = Field(
        description="Externally provided ID for the subscription. Must be unique.",
        max_length=255,
    )
    name: str | None = Field(
        default=None,
        description="Human readable name of the subscription.",
    )
    description: str | None = Field(default=None, description="A description of the subscription.")
    data_set_external_id: str | None = Field(
        default=None,
        description="The dataSet Id for the item.",
    )
    partition_count: int | None = Field(
        default=None,
        ge=1,
        le=100,
        description="The maximum effective parallelism of this subscription (the number of clients that can read from it concurrently) will be limited to this number, but a higher partition count will cause a higher time overhead.",
    )
    time_series_ids: list[str] | None = Field(
        None,
        description="List of (external) ids of time series that this subscription will listen to. Not compatible with filter.",
    )
    instance_ids: list[InstanceId] | None = Field(
        None,
        description="List of instance ids of time series that this subscription will listen to. Not compatible with filter.",
    )

    filter: dict[str, JsonValue] | None = Field(
        default=None,
        description="A filter Domain Specific Language (DSL) used to create advanced filter queries. Not compatible with time_series_ids or instance_ids.",
    )

    @model_validator(mode="after")
    def check_subscription_targets(self) -> Self:
        time_series_ids = self.time_series_ids or []
        instance_ids = self.instance_ids or []

        total = len(time_series_ids) + len(instance_ids)
        if total > _SUBSCRIPTION_MAX_TOTAL_TIMESERIES:
            raise ValueError(
                f"The total number of time_series_ids and instance_ids cannot exceed {_SUBSCRIPTION_MAX_TOTAL_TIMESERIES}."
            )

        if self.filter is not None and total > 0:
            raise ValueError("Cannot set both filter and time_series_ids/instance_ids.")
        return self



from abc import ABC, abstractmethod

from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses


class GrandParent(ABC):
    @abstractmethod
    def method(self) -> None:
        pass


class Parent(GrandParent, ABC): ...


class Parent2(GrandParent):
    def method(self) -> None:
        print("Parent2 method implementation")


class Child(Parent):
    def method(self) -> None:
        print("Child method implementation")


class Child2(Parent):
    def method(self) -> None:
        print("Child2 method implementation")


class Child3(Child, Parent2):
    def method(self) -> None:
        print("Child3 method implementation")


class Mixin1(ABC):
    pass


class Mixin2(ABC):
    pass


class MultiChild(Mixin1, Mixin2, GrandParent):
    def method(self) -> None:
        pass


class TestGetConcreteSubclasses:
    def test_get_concrete_subclasses(self) -> None:
        subclasses = get_concrete_subclasses(GrandParent)

        assert set(subclasses) == {Parent2, Child, Child2, Child3, MultiChild}

    def test_multiple_inheritance(self) -> None:
        subclasses = get_concrete_subclasses(GrandParent)
        assert len(subclasses) == len(set(subclasses)), "Subclasses should be unique"



import pytest

from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.utils.cli_args import parse_view_str


class TestParseViewStr:
    def test_valid_str(self) -> None:
        view_str = "my_space:MyView/v2"
        view_id = parse_view_str(view_str)
        assert view_id.space == "my_space"
        assert view_id.external_id == "MyView"
        assert view_id.version == "v2"

    @pytest.mark.parametrize(
        "invalid_str, expected_message",
        [
            (
                "invalidformat",
                "Invalid view string format: 'invalidformat'. Expected format 'space:externalId/version'.",
            ),
            ("spaceonly:", "Invalid view string format: 'spaceonly:'. Expected format 'space:externalId/version'."),
            (":/v1", "Invalid view string format: ':/v1'. Expected format 'space:externalId/version'."),
            (
                "space:externalId",
                "Invalid view string format: 'space:externalId'. Expected format 'space:externalId/version'.",
            ),
            (
                "space:externalId/",
                "Invalid view string format: 'space:externalId/'. Expected format 'space:externalId/version'.",
            ),
        ],
    )
    def test_invalid_str(self, invalid_str: str, expected_message: str) -> None:
        with pytest.raises(ToolkitValueError) as exc_info:
            parse_view_str(invalid_str)
        assert str(exc_info.value) == expected_message



from abc import ABC, abstractmethod
from functools import lru_cache
from typing import ClassVar, Generic, Literal, TypeVar

from cognite.client.data_classes import (
    AssetFilter,
    EventFilter,
    FileMetadataFilter,
    SequenceFilter,
    TimeSeriesFilter,
    Transformation,
    filters,
)
from cognite.client.data_classes.assets import AssetProperty
from cognite.client.data_classes.documents import SourceFileProperty
from cognite.client.data_classes.events import EventProperty
from cognite.client.data_classes.sequences import SequenceProperty
from cognite.client.data_classes.time_series import TimeSeriesProperty

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.exceptions import ToolkitMissingResourceError
from cognite_toolkit._cdf_tk.utils.cdf import (
    label_aggregate_count,
    label_count,
    metadata_key_counts,
    relationship_aggregate_count,
)
from cognite_toolkit._cdf_tk.utils.sql_parser import SQLParser

T_CogniteFilter = TypeVar(
    "T_CogniteFilter",
    bound=AssetFilter | EventFilter | FileMetadataFilter | TimeSeriesFilter | SequenceFilter,
    contravariant=True,
)


class AssetCentricAggregator(ABC):
    _transformation_destination: ClassVar[tuple[str, ...]]

    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    @property
    @abstractmethod
    def display_name(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def count(
        self, hierarchy: str | list[str] | None = None, data_set_external_id: str | list[str] | None = None
    ) -> int:
        raise NotImplementedError

    def transformation_count(self) -> int:
        """Returns the number of transformations associated with the resource."""
        transformation_count = 0
        for destination in self._transformation_destination:
            for chunk in self.client.transformations(chunk_size=1000, destination_type=destination, limit=None):
                transformation_count += len(chunk)
        return transformation_count

    @abstractmethod
    def used_data_sets(self, hierarchy: str | None = None) -> list[str]:
        """Returns a list of data sets used by the resource."""
        raise NotImplementedError

    def used_transformations(self, data_set_external_ids: list[str]) -> list[Transformation]:
        """Returns a list of transformations used by the resource."""
        data_set_ids = self.client.lookup.data_sets.id(data_set_external_ids, allow_empty=True)
        found_transformations: list[Transformation] = []
        for destination in self._transformation_destination:
            for chunk in self.client.transformations(chunk_size=1000, destination_type=destination, limit=None):
                for transformation in chunk:
                    if SQLParser(transformation.query or "", operation="profiling").is_using_data_set(
                        data_set_ids, data_set_external_ids
                    ):
                        found_transformations.append(transformation)
        return found_transformations

    @staticmethod
    def _to_unique_int_list(results: list) -> list[int]:
        """Converts a list of results to a unique list of integers.

        This method does the following:
        * Converts each item in the results to an integer, if possible.
        * Filters out None.
        * Removes duplicates

        This is used as the aggregation results are inconsistently implemented for the different resources,
        when aggregating dataSetIds, Sequences, TimeSeries, and Files return a list of strings, while
        Assets and Events return a list of integers. In addition, the files aggregation can return
        duplicated values.
        """
        seen: set[int] = set()
        ids: list[int] = []
        for id_ in results:
            if isinstance(id_, int) and id_ not in seen:
                ids.append(id_)
                seen.add(id_)
            try:
                int_id = int(id_)
            except (ValueError, TypeError):
                continue
            if int_id not in seen:
                ids.append(int_id)
                seen.add(int_id)
        return ids

    def _to_dataset_id(self, data_set_external_id: str | list[str] | None) -> list[int] | None:
        """Converts data set external IDs to data set IDs."""
        dataset_id: list[int] | None = None
        if data_set_external_id is not None:
            if isinstance(data_set_external_id, str):
                data_set_external_id = [data_set_external_id]
            dataset_id = self.client.lookup.data_sets.id(data_set_external_id, allow_empty=False)
        return dataset_id


class MetadataAggregator(AssetCentricAggregator, ABC, Generic[T_CogniteFilter]):
    filter_cls: type[T_CogniteFilter]

    def __init__(
        self, client: ToolkitClient, resource_name: Literal["assets", "events", "files", "timeseries", "sequences"]
    ) -> None:
        super().__init__(client)
        self.resource_name = resource_name

    def metadata_key_count(
        self, hierarchy: str | list[str] | None = None, data_sets: str | list[str] | None = None
    ) -> int:
        """Returns the number of metadata keys used by the resource."""
        return len(self.used_metadata_keys(hierarchy=hierarchy, data_sets=data_sets))

    def used_metadata_keys(
        self, hierarchy: str | list[str] | None = None, data_sets: str | list[str] | None = None
    ) -> list[tuple[str, int]]:
        """Returns a list of metadata keys and their counts."""
        hierarchy_ids, data_set_ids = self._lookup_hierarchy_data_set_pair(
            hierarchy, data_sets, operation="find metadata keys"
        )
        return self._used_metadata_keys(hierarchy=hierarchy_ids, data_sets=data_set_ids)

    @lru_cache(maxsize=1)
    def _used_metadata_keys(
        self, hierarchy: tuple[int, ...] | None = None, data_sets: tuple[int, ...] | None = None
    ) -> list[tuple[str, int]]:
        return metadata_key_counts(
            self.client,
            self.resource_name,
            hierarchies=list(hierarchy) if hierarchy else None,
            data_sets=list(data_sets) if data_sets else None,
        )

    def _lookup_hierarchy_data_set_pair(
        self, hierarchy: str | list[str] | None, data_sets: str | list[str] | None, operation: str
    ) -> tuple[tuple[int, ...] | None, tuple[int, ...] | None]:
        """Returns a tuple of hierarchy and data sets."""
        hierarchy_ids: tuple[int, ...] | None = None
        if isinstance(hierarchy, str):
            asset_id = self.client.lookup.assets.id(external_id=hierarchy, allow_empty=False)
            if asset_id is None:
                raise ToolkitMissingResourceError(f"Cannot {operation}. Asset with external ID {hierarchy!r} not found")
            hierarchy_ids = (asset_id,)
        elif isinstance(hierarchy, list) and all(isinstance(item, str) for item in hierarchy):
            asset_ids = self.client.lookup.assets.id(external_id=hierarchy, allow_empty=False)
            if len(asset_ids) != len(hierarchy):
                missing = set(hierarchy) - set(
                    self.client.lookup.assets.external_id([id_ for id_ in asset_ids if id_ is not None])
                )
                raise ToolkitMissingResourceError(
                    f"Cannot {operation}. Assets with external IDs {sorted(missing)!r} not found"
                )
            hierarchy_ids = tuple(sorted(asset_ids))

        data_set_ids: tuple[int, ...] | None = None
        if isinstance(data_sets, str):
            data_set_id = self.client.lookup.data_sets.id(external_id=data_sets, allow_empty=False)
            if data_set_id is None:
                raise ToolkitMissingResourceError(
                    f"Cannot {operation}. Data set with external ID {data_sets!r} not found"
                )
            data_set_ids = (data_set_id,)
        elif isinstance(data_sets, list) and all(isinstance(item, str) for item in data_sets):
            data_set_ids_list = self.client.lookup.data_sets.id(external_id=data_sets, allow_empty=False)
            if len(data_set_ids_list) != len(data_sets):
                missing = set(data_sets) - set(
                    self.client.lookup.data_sets.external_id([id_ for id_ in data_set_ids_list if id_ is not None])
                )
                raise ToolkitMissingResourceError(
                    f"Cannot {operation}. Data sets with external IDs {sorted(missing)!r} not found"
                )
            data_set_ids = tuple(sorted(data_set_ids_list))

        return hierarchy_ids, data_set_ids

    @classmethod
    def create_filter(
        cls,
        hierarchy: str | list[str] | tuple[str, ...] | None = None,
        data_set_external_id: str | list[str] | tuple[str, ...] | None = None,
    ) -> T_CogniteFilter | None:
        """Creates a filter for the resource based on hierarchy and data set external ID."""
        if cls._is_empty(hierarchy) and cls._is_empty(data_set_external_id):
            return None
        asset_subtree_ids: list[dict[str, str]] | None = None
        if isinstance(hierarchy, str):
            asset_subtree_ids = [{"externalId": hierarchy}]
        elif isinstance(hierarchy, list | tuple) and hierarchy:
            asset_subtree_ids = [{"externalId": item} for item in hierarchy]
        data_set_ids: list[dict[str, str]] | None = None
        if isinstance(data_set_external_id, str):
            data_set_ids = [{"externalId": data_set_external_id}]
        elif isinstance(data_set_external_id, list | tuple) and data_set_external_id:
            data_set_ids = [{"externalId": item} for item in data_set_external_id]

        # MyPy fails to understand that filter_cls() produce a T_CogniteFilter
        return cls.filter_cls(asset_subtree_ids=asset_subtree_ids, data_set_ids=data_set_ids)  # type: ignore[return-value]

    @classmethod
    def _is_empty(cls, items: str | list[str] | tuple[str, ...] | None) -> bool:
        """Checks if the provided items are empty."""
        if items is None:
            return True
        if isinstance(items, list | tuple):
            return not items
        return False


class LabelAggregator(MetadataAggregator, ABC, Generic[T_CogniteFilter]):
    def label_count(self, hierarchy: str | list[str] | None = None, data_sets: str | list[str] | None = None) -> int:
        """Returns the number of labels used by the resource."""
        return len(self.used_labels(hierarchy=hierarchy, data_sets=data_sets))

    def used_labels(
        self, hierarchy: str | list[str] | None = None, data_sets: str | list[str] | None = None
    ) -> list[tuple[str, int]]:
        """Returns a list of labels and their counts."""
        hierarchy_ids, data_set_ids = self._lookup_hierarchy_data_set_pair(
            hierarchy, data_sets, operation="find labels"
        )
        return self._used_labels(hierarchy=hierarchy_ids, data_sets=data_set_ids)

    @lru_cache(maxsize=1)
    def _used_labels(
        self, hierarchy: tuple[int, ...] | None = None, data_sets: tuple[int, ...] | None = None
    ) -> list[tuple[str, int]]:
        """Returns a list of labels and their counts."""
        return label_count(
            self.client,
            self.resource_name,
            hierarchies=list(hierarchy) if hierarchy else None,
            data_sets=list(data_sets) if data_sets else None,
        )


class AssetAggregator(LabelAggregator[AssetFilter]):
    _transformation_destination = ("assets", "asset_hierarchy")
    filter_cls = AssetFilter

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "assets")

    @property
    def display_name(self) -> str:
        return "Assets"

    def count(
        self,
        hierarchy: str | list[str] | tuple[str, ...] | None = None,
        data_set_external_id: str | list[str] | tuple[str, ...] | None = None,
    ) -> int:
        return self.client.assets.aggregate_count(filter=self.create_filter(hierarchy, data_set_external_id))

    def used_data_sets(self, hierarchy: str | None = None) -> list[str]:
        """Returns a list of data sets used by the resource."""
        results = self.client.assets.aggregate_unique_values(
            AssetProperty.data_set_id, filter=self.create_filter(hierarchy)
        )
        return self.client.lookup.data_sets.external_id([id_ for id_ in results.unique if isinstance(id_, int)])


class EventAggregator(MetadataAggregator[EventFilter]):
    _transformation_destination = ("events",)
    filter_cls = EventFilter

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "events")

    @property
    def display_name(self) -> str:
        return "Events"

    def count(
        self, hierarchy: str | list[str] | None = None, data_set_external_id: str | list[str] | None = None
    ) -> int:
        return self.client.events.aggregate_count(filter=self.create_filter(hierarchy, data_set_external_id))

    def used_data_sets(self, hierarchy: str | None = None) -> list[str]:
        """Returns a list of data sets used by the resource."""
        results = self.client.events.aggregate_unique_values(
            property=EventProperty.data_set_id, filter=self.create_filter(hierarchy)
        )
        return self.client.lookup.data_sets.external_id([id_ for id_ in results.unique if isinstance(id_, int)])


class FileAggregator(LabelAggregator[FileMetadataFilter]):
    _transformation_destination = ("files",)
    filter_cls = FileMetadataFilter

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "files")

    @property
    def display_name(self) -> str:
        return "Files"

    def count(
        self, hierarchy: str | list[str] | None = None, data_set_external_id: str | list[str] | None = None
    ) -> int:
        response = self.client.files.aggregate(filter=self.create_filter(hierarchy, data_set_external_id))
        if response:
            return response[0].count
        else:
            return 0

    def used_data_sets(self, hierarchy: str | None = None) -> list[str]:
        """Returns a list of data sets used by the resource."""
        filter_: filters.Filter | None = None
        if hierarchy is not None:
            filter_ = filters.InAssetSubtree("assetExternalIds", [hierarchy])
        results = self.client.documents.aggregate_unique_values(
            property=SourceFileProperty.data_set_id, filter=filter_, limit=1000
        )
        ids = self._to_unique_int_list(results.unique)
        return self.client.lookup.data_sets.external_id(list(ids))


class TimeSeriesAggregator(MetadataAggregator[TimeSeriesFilter]):
    _transformation_destination = ("timeseries",)
    filter_cls = TimeSeriesFilter

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "timeseries")

    @property
    def display_name(self) -> str:
        return "TimeSeries"

    def count(
        self, hierarchy: str | list[str] | None = None, data_set_external_id: str | list[str] | None = None
    ) -> int:
        return self.client.time_series.aggregate_count(filter=self.create_filter(hierarchy, data_set_external_id))

    def used_data_sets(self, hierarchy: str | None = None) -> list[str]:
        """Returns a list of data sets used by the resource."""
        results = self.client.time_series.aggregate_unique_values(
            property=TimeSeriesProperty.data_set_id, filter=self.create_filter(hierarchy)
        )
        ids = self._to_unique_int_list(results.unique)
        return self.client.lookup.data_sets.external_id(ids)


class SequenceAggregator(MetadataAggregator):
    _transformation_destination = ("sequences",)
    filter_cls = SequenceFilter

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "sequences")

    @property
    def display_name(self) -> str:
        return "Sequences"

    def count(
        self, hierarchy: str | list[str] | None = None, data_set_external_id: str | list[str] | None = None
    ) -> int:
        return self.client.sequences.aggregate_count(filter=self.create_filter(hierarchy, data_set_external_id))

    def used_data_sets(self, hierarchy: str | None = None) -> list[str]:
        """Returns a list of data sets used by the resource."""
        results = self.client.sequences.aggregate_unique_values(
            property=SequenceProperty.data_set_id, filter=self.create_filter(hierarchy)
        )
        ids = self._to_unique_int_list(results.unique)
        return self.client.lookup.data_sets.external_id(ids)


class RelationshipAggregator(AssetCentricAggregator):
    _transformation_destination = ("relationships",)

    @property
    def display_name(self) -> str:
        return "Relationships"

    def count(
        self, hierarchy: str | list[str] | None = None, data_set_external_id: str | list[str] | None = None
    ) -> int:
        if hierarchy is not None:
            raise NotImplementedError()
        dataset_id = self._to_dataset_id(data_set_external_id)
        results = relationship_aggregate_count(self.client, dataset_id)
        return sum(result.count for result in results)

    def used_data_sets(self, hierarchy: str | None = None) -> list[str]:
        raise NotImplementedError()


class LabelCountAggregator(AssetCentricAggregator):
    _transformation_destination = ("labels",)

    @property
    def display_name(self) -> str:
        return "Labels"

    def count(
        self, hierarchy: str | list[str] | None = None, data_set_external_id: str | list[str] | None = None
    ) -> int:
        if hierarchy is not None:
            raise NotImplementedError()
        data_set_id = self._to_dataset_id(data_set_external_id)
        return label_aggregate_count(self.client, data_set_id)

    def used_data_sets(self, hierarchy: str | None = None) -> list[str]:
        raise NotImplementedError()



from functools import cached_property

from cognite.client import CogniteClient
from cognite.client.data_classes.capabilities import _CAPABILITY_CLASS_BY_NAME, AllScope, Capability
from cognite.client.data_classes.iam import TokenInspection

from cognite_toolkit._cdf_tk.client.data_classes.capabilities import scope_intersection, scope_union
from cognite_toolkit._cdf_tk.utils import humanize_collection

_ACL_CLASS_BY_CLASS_NAME = {cap.__name__: cap for cap in _CAPABILITY_CLASS_BY_NAME.values()}


class TokenAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    @cached_property
    def token(self) -> TokenInspection:
        return self._client.iam.token.inspect()

    def get_scope(self, actions: list[Capability.Action]) -> list[Capability.Scope] | None:
        """Gets the scopes for the given actions and ACL class from the token.

        Args:
            actions:  List of actions to get scopes for.

        Returns:
            A list of scopes that apply to all the actions, or None if no scopes are found.
        """
        acl_cls = self._get_acl_class(actions)
        # First, we find the union of all scopes for a given action,
        # then we find the intersection of those scopes across all actions.
        scope_union_by_action_by_scope_cls = self._union_by_action_by_scope_cls(acl_cls, set(actions))
        if not scope_union_by_action_by_scope_cls:
            return None
        return self._intersection_over_actions(scope_union_by_action_by_scope_cls)

    @staticmethod
    def _get_acl_class(actions: list[Capability.Action]) -> type[Capability]:
        """Returns the ACL class that matches the given actions."""
        classes: set[type[Capability]] = set()
        for action in actions:
            action_class_name = type(action).__qualname__.split(".")[0]
            if action_class_name not in _ACL_CLASS_BY_CLASS_NAME:
                raise ValueError(f"Unknown action class: {action_class_name}")
            classes.add(_ACL_CLASS_BY_CLASS_NAME[action_class_name])
        if len(classes) > 1:
            raise ValueError(
                f"Actions belong to multiple ACL classes: {humanize_collection([cls.__name__ for cls in classes])}."
            )
        if len(classes) == 0:
            raise ValueError("No actions provided to get_scope")
        return next(iter(classes))

    @staticmethod
    def _intersection_over_actions(
        scopes_by_action: dict[Capability.Action, dict[type[Capability.Scope], Capability.Scope]],
    ) -> list[Capability.Scope] | None:
        all_scope_actions = {
            action for action, scopes_by_cls in scopes_by_action.items() if AllScope in scopes_by_cls.keys()
        }
        if len(all_scope_actions) == len(scopes_by_action):
            # Special case: if AllScope is in all actions, return it directly
            return [AllScope()]
        intersection_scopes = set.intersection(
            *[
                set(scopes_by_cls.keys())
                for action, scopes_by_cls in scopes_by_action.items()
                if action not in all_scope_actions
            ]
        )
        if not intersection_scopes:
            return None
        output_scopes: list[Capability.Scope] = []
        for scope_cls in sorted(intersection_scopes, key=lambda cls: cls.__name__):
            scopes = [
                scopes_by_cls[scope_cls] for scopes_by_cls in scopes_by_action.values() if scope_cls in scopes_by_cls
            ]
            if len(scopes) == 0:
                # Should not happen, as we already checked for intersection_scopes
                continue
            elif len(scopes) == 1:
                output_scopes.append(scopes[0])
            else:
                intersection: Capability.Scope | None = scopes[0]
                for scope in scopes[1:]:
                    intersection = scope_intersection(scope, intersection)
                    if intersection is None:
                        break
                if intersection is None:
                    # If the intersection is None, it means the scopes do not intersect
                    continue
                output_scopes.append(intersection)
        return output_scopes or None

    def _union_by_action_by_scope_cls(
        self, acl_cls: type[Capability], actions: set[Capability.Action]
    ) -> dict[Capability.Action, dict[type[Capability.Scope], Capability.Scope]]:
        scopes_by_action: dict[Capability.Action, dict[type[Capability.Scope], Capability.Scope]] = {}
        for action in actions:
            scopes_by_action[action] = {}
        for project_capability in self.token.capabilities:
            capability = project_capability.capability
            if not isinstance(capability, acl_cls):
                continue
            for action in capability.actions:
                if action in actions:
                    scopes_by_cls = scopes_by_action[action]
                    scope_cls = type(capability.scope)
                    scopes_by_cls[scope_cls] = scope_union(capability.scope, scopes_by_cls.get(scope_cls))
        return scopes_by_action



from typing import Literal

from pydantic import Field

from .base import BaseModelResource, ToolkitResource


class RawTable(BaseModelResource):
    db_name: str = Field(description="Database name", min_length=1)
    table_name: str = Field(description="Table name", min_length=1)


class Contact(BaseModelResource):
    name: str | None = Field(None, description="Contact name")
    email: str | None = Field(None, description="Contact email", min_length=1, max_length=254)
    role: str | None = Field(None, description="Contact role")
    send_notification: bool | None = Field(None, description="True, if contact receives email notifications")


class NotificationConfig(BaseModelResource):
    allowed_not_seen_range_in_minutes: int | None = Field(
        None,
        ge=0,
        description="Notifications configuration value. Time in minutes to pass without any Run. Null if extraction pipeline is not checked.",
    )


class ExtractionPipelineYAML(ToolkitResource):
    external_id: str = Field(
        description="External Id provided by client. Should be unique within the project.", min_length=1, max_length=255
    )
    name: str = Field(description="Name of the extraction pipeline.", min_length=1, max_length=140)
    description: str | None = Field(None, description="Description of the extraction pipeline.", max_length=500)
    data_set_external_id: str = Field(description="The external id of the dataset this extraction pipeline belongs to.")
    raw_tables: list[RawTable] | None = Field(None, description="Raw tables")
    schedule: Literal["On trigger", "Continuous"] | str | None = Field(
        None,
        description="Possible values: “On trigger”, “Continuous” or cron expression. If empty then null.",
    )
    contacts: list[Contact] | None = Field(None, description="Contacts list.")
    metadata: dict[str, str] | None = Field(
        None,
        description="Custom, application specific metadata. String key -> String value. Limits: Key are at most 128 bytes. Values are at most 10240 bytes. Up to 256 key-value pairs. Total size is at most 10240.",
    )
    source: str | None = Field(None, description="Source for Extraction Pipeline", max_length=255)
    documentation: str | None = Field(
        None, description="Documentation text field, supports Markdown for text formatting.", max_length=10000
    )
    notification_config: NotificationConfig | None = Field(
        None, description="Notification configuration for the extraction pipeline."
    )
    created_by: str | None = Field(
        None, description="Extraction Pipeline creator. Usually user email is expected here."
    )



from cognite.client.data_classes import SequenceWrite
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.cruds import SequenceCRUD


class TestSequenceLoader:
    def test_load_timeseries_ref_not_yet_deployed(self) -> None:
        actual_is_dry_run: list[bool] = []

        def mock_id_lookup(
            external_id: str | SequenceNotStr[str], is_dry_run: bool = False, allow_empty: bool = False
        ) -> int | list[int]:
            actual_is_dry_run.append(is_dry_run)
            return -1 if isinstance(external_id, str) else [-1] * len(external_id)

        with monkeypatch_toolkit_client() as client:
            client.lookup.data_sets.id.side_effect = mock_id_lookup
            client.lookup.assets.id.side_effect = mock_id_lookup
            client.lookup.security_categories.id.side_effect = mock_id_lookup
            loader = SequenceCRUD.create_loader(client)

        resource = {
            "externalId": "mySequence",
            "columns": [
                {"externalId": "myColumn1"},
                {"externalId": "myColumn2"},
            ],
            "assetExternalId": "my_asset",
            "dataSetExternalId": "my_dataset",
        }

        seq = loader.load_resource(resource.copy(), is_dry_run=True)
        _ = loader.load_resource(resource.copy(), is_dry_run=False)

        assert isinstance(seq, SequenceWrite)
        assert seq.asset_id == -1
        assert seq.data_set_id == -1
        assert actual_is_dry_run == [True] * 2 + [False] * 2



import pytest
import requests

from cognite_toolkit._cdf_tk.constants import URL


@pytest.mark.parametrize(
    "url, name",
    [pytest.param(url, name, id=name) for name, url in vars(URL).items() if not name.startswith("_")],
)
def test_url_returns_200(url: str, name: str) -> None:
    assert requests.get(url).status_code == 200, f"Failed to get a 200 response from the URL.{name}."



from random import randint

import pytest

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.search_config import (
    SearchConfig,
    SearchConfigWrite,
    ViewId,
)

SEARCH_CONFIG_NAME = "Search Config Name"


@pytest.fixture(scope="session")
def existing_search_config(toolkit_client: ToolkitClient) -> SearchConfig:
    view = ViewId(external_id="CogniteTimeSeries", space="cdf_cdm")
    search_config = SearchConfigWrite(view=view, use_as_name=SEARCH_CONFIG_NAME)

    configs = toolkit_client.search.configurations.list()
    if configs:
        for config in configs:
            if config.view == view:
                return config

    created = toolkit_client.search.configurations.upsert(search_config)
    return created


class TestSearchConfigAPI:
    def test_update_search_configuration(
        self, toolkit_client: ToolkitClient, existing_search_config: SearchConfig
    ) -> None:
        view = existing_search_config.view
        test_description = f"Test description Update {randint(1000, 9999)}"
        test_name = SEARCH_CONFIG_NAME
        search_config = SearchConfigWrite(
            id=existing_search_config.id,
            view=view,
            use_as_description=test_description,
            use_as_name=test_name,
        )

        updated = toolkit_client.search.configurations.upsert(search_config)
        assert isinstance(updated, SearchConfig)
        assert updated.id == existing_search_config.id
        assert updated.view.external_id == view.external_id
        assert updated.view.space == view.space
        assert updated.use_as_name == SEARCH_CONFIG_NAME
        assert updated.use_as_description == test_description

        configs = toolkit_client.search.configurations.list()
        found_config = next((c for c in configs if c.id == updated.id), None)
        assert found_config is not None
        assert found_config.use_as_name == SEARCH_CONFIG_NAME



from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.containers import BTreeIndex

SPACE = dm.SpaceApply(
    "cognite_migration", description="Space for the asset-centric to data modeling migration", name="cdf_migration"
)

RESOURCE_VIEW_MAPPING = dm.ContainerApply(
    space=SPACE.space,
    external_id="ResourceViewMapping",
    used_for="node",
    properties={
        "resourceType": dm.ContainerProperty(
            type=dm.Text(max_text_size=255),
            nullable=False,
        ),
        "viewId": dm.ContainerProperty(
            type=dm.data_types.Json(),
            nullable=False,
        ),
        "propertyMapping": dm.ContainerProperty(
            type=dm.data_types.Json(),
            nullable=False,
        ),
    },
)

INSTANCE_SOURCE_CONTAINER = dm.ContainerApply(
    space=SPACE.space,
    external_id="InstanceSource",
    used_for="node",
    properties={
        "resourceType": dm.ContainerProperty(
            type=dm.data_types.Enum(
                values={
                    "timeseries": dm.data_types.EnumValue(),
                    "asset": dm.data_types.EnumValue(),
                    "file": dm.data_types.EnumValue(),
                    "event": dm.data_types.EnumValue(),
                    "sequence": dm.data_types.EnumValue(),
                }
            ),
            nullable=False,
        ),
        "id": dm.ContainerProperty(
            type=dm.data_types.Int64(),
            nullable=False,
        ),
        "dataSetId": dm.ContainerProperty(
            type=dm.data_types.Int64(),
            nullable=True,
        ),
        "classicExternalId": dm.ContainerProperty(
            type=dm.data_types.Text(),
            nullable=True,
        ),
        "preferredConsumerViewId": dm.ContainerProperty(
            type=dm.data_types.Json(),
            nullable=True,
        ),
        "resourceViewMapping": dm.ContainerProperty(
            type=dm.data_types.DirectRelation(container=RESOURCE_VIEW_MAPPING.as_id()), nullable=True
        ),
    },
    indexes={
        "id": BTreeIndex(["id"], cursorable=True),
        "resourceType": BTreeIndex(["resourceType", "id"], cursorable=False),
    },
)

CREATED_SOURCE_SYSTEM = dm.ContainerApply(
    space=SPACE.space,
    external_id="CreatedSourceSystem",
    used_for="node",
    properties={
        "source": dm.ContainerProperty(
            type=dm.Text(max_text_size=128),
            nullable=False,
        )
    },
    constraints={"sourceUnique": dm.UniquenessConstraint(["source"])},
    indexes={
        "source": BTreeIndex(["source"], cursorable=True),
    },
)

SPACE_SOURCE = dm.ContainerApply(
    space=SPACE.space,
    external_id="SpaceSource",
    used_for="node",
    properties={
        "instanceSpace": dm.ContainerProperty(
            type=dm.Text(max_text_size=64),
            nullable=False,
        ),
        "dataSetId": dm.ContainerProperty(
            type=dm.data_types.Int64(),
            nullable=False,
        ),
        "dataSetExternalId": dm.ContainerProperty(
            type=dm.Text(max_text_size=256),
            nullable=True,
        ),
    },
    indexes={
        "space": BTreeIndex(["instanceSpace"], cursorable=True),
        "dataSetId": BTreeIndex(["dataSetId"], cursorable=True),
        "dataSetExternalId": BTreeIndex(["dataSetExternalId"], cursorable=True),
    },
    constraints={
        "dataSetIdUnique": dm.UniquenessConstraint(["dataSetId"]),
    },
)

CONTAINERS = [RESOURCE_VIEW_MAPPING, INSTANCE_SOURCE_CONTAINER, CREATED_SOURCE_SYSTEM, SPACE_SOURCE]

RESOURCE_VIEW_MAPPING_VIEW = dm.ViewApply(
    space=SPACE.space,
    external_id="ResourceViewMapping",
    version="v1",
    name="ResourceViewMapping",
    description="The mapping from asset-centric resources to data modeling view.",
    properties={
        "resourceType": dm.MappedPropertyApply(
            container=RESOURCE_VIEW_MAPPING.as_id(),
            container_property_identifier="resourceType",
        ),
        "viewId": dm.MappedPropertyApply(
            container=RESOURCE_VIEW_MAPPING.as_id(),
            container_property_identifier="viewId",
        ),
        "propertyMapping": dm.MappedPropertyApply(
            container=RESOURCE_VIEW_MAPPING.as_id(),
            container_property_identifier="propertyMapping",
        ),
    },
)

INSTANCE_SOURCE_VIEW = dm.ViewApply(
    space=SPACE.space,
    external_id="InstanceSource",
    version="v1",
    name="InstanceSource",
    description="The source of the instance in asset-centric resources.",
    properties={
        "resourceType": dm.MappedPropertyApply(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="resourceType",
        ),
        "id": dm.MappedPropertyApply(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="id",
        ),
        "dataSetId": dm.MappedPropertyApply(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="dataSetId",
        ),
        "classicExternalId": dm.MappedPropertyApply(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="classicExternalId",
        ),
        "preferredConsumerViewId": dm.MappedPropertyApply(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="preferredConsumerViewId",
        ),
        "resourceViewMapping": dm.MappedPropertyApply(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="resourceViewMapping",
            source=RESOURCE_VIEW_MAPPING_VIEW.as_id(),
        ),
    },
)

CREATED_SOURCE_SYSTEM_VIEW = dm.ViewApply(
    space=SPACE.space,
    external_id="CreatedSourceSystem",
    version="v1",
    name="CreatedSourceSystem",
    description="The source string the SourceSystem was created from.",
    properties={
        "source": dm.MappedPropertyApply(
            container=CREATED_SOURCE_SYSTEM.as_id(),
            container_property_identifier="source",
        ),
    },
)

SPACE_SOURCE_VIEW = dm.ViewApply(
    space=SPACE.space,
    external_id="SpaceSource",
    version="v1",
    name="SpaceSource",
    description="The mapping from CDF spaces to data sets.",
    properties={
        "instanceSpace": dm.MappedPropertyApply(
            container=SPACE_SOURCE.as_id(),
            container_property_identifier="instanceSpace",
            description="The identifier of the created instances space.",
        ),
        "dataSetId": dm.MappedPropertyApply(
            container=SPACE_SOURCE.as_id(),
            container_property_identifier="dataSetId",
            description="The dataSetId the space was created from.",
        ),
        "dataSetExternalId": dm.MappedPropertyApply(
            container=SPACE_SOURCE.as_id(),
            container_property_identifier="dataSetExternalId",
            description="The externalId of the dataSet (if present) the space was created from.",
        ),
    },
)

INSTANCE_SOURCE_VIEW_ID = INSTANCE_SOURCE_VIEW.as_id()
RESOURCE_VIEW_MAPPING_VIEW_ID = RESOURCE_VIEW_MAPPING_VIEW.as_id()
CREATED_SOURCE_SYSTEM_VIEW_ID = CREATED_SOURCE_SYSTEM_VIEW.as_id()
SPACE_SOURCE_VIEW_ID = SPACE_SOURCE_VIEW.as_id()

VIEWS = [RESOURCE_VIEW_MAPPING_VIEW, INSTANCE_SOURCE_VIEW, CREATED_SOURCE_SYSTEM_VIEW, SPACE_SOURCE_VIEW]

COGNITE_MIGRATION_MODEL = dm.DataModelApply(
    space=SPACE.space,
    external_id="CogniteMigration",
    version="v1",
    name="CDF Migration Model",
    description="Data model for migrating asset-centric resources to data modeling resources in CDF.",
    views=[INSTANCE_SOURCE_VIEW_ID, RESOURCE_VIEW_MAPPING_VIEW_ID, CREATED_SOURCE_SYSTEM_VIEW_ID, SPACE_SOURCE_VIEW_ID],
)

MODEL_ID = COGNITE_MIGRATION_MODEL.as_id()



from __future__ import annotations

from pathlib import Path

from cognite_toolkit._cdf_tk.constants import COMPLETE_ORG
from cognite_toolkit._cdf_tk.data_classes import ModuleDirectories, ModuleLocation
from tests import data


class TestModuleDirectories:
    def test_load(self) -> None:
        cognite_modules = Path(COMPLETE_ORG)
        expected = ModuleDirectories(
            [
                ModuleLocation(
                    dir=cognite_modules / Path("a_module"),
                    source_absolute_path=data.PROJECT_FOR_TEST,
                    is_selected=True,
                    source_paths=[],
                ),
                ModuleLocation(
                    dir=cognite_modules / Path("another_module"),
                    source_absolute_path=data.PROJECT_FOR_TEST,
                    is_selected=True,
                    source_paths=[],
                ),
                ModuleLocation(
                    dir=cognite_modules / Path("parent_module") / "child_module",
                    source_absolute_path=data.PROJECT_FOR_TEST,
                    is_selected=True,
                    source_paths=[],
                ),
            ]
        )
        actual = ModuleDirectories.load(data.PROJECT_FOR_TEST, {("modules",)})

        assert actual == expected



import sys
from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class BaseModelObject(BaseModel):
    """Base class for all object. This includes resources and nested objects."""

    # We allow extra fields to support forward compatibility.
    model_config = ConfigDict(alias_generator=to_camel, extra="allow")

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the resource to a dictionary.

        This is the default serialization method for request resources.
        """
        return self.model_dump(mode="json", by_alias=camel_case)

    @classmethod
    def _load(cls, resource: dict[str, Any]) -> "Self":
        """Load method to match CogniteResource signature."""
        return cls.model_validate(resource)


class RequestResource(BaseModelObject): ...


T_RequestResource = TypeVar("T_RequestResource", bound=RequestResource)


class ResponseResource(BaseModelObject, Generic[T_RequestResource], ABC):
    @abstractmethod
    def as_request_resource(self) -> T_RequestResource:
        """Convert the response resource to a request resource."""
        ...


class Identifier(BaseModel):
    """Base class for all identifier classes."""

    model_config = ConfigDict(alias_generator=to_camel, extra="ignore", populate_by_name=True, frozen=True)

    def dump(self, include_type: bool = True) -> dict[str, Any]:
        """Dump the identifier to a dictionary.

        Args:
            include_type (bool): Whether to include the type of the identifier in the output.

        Returns:
            dict[str, Any]: The dumped identifier.
        """
        return self.model_dump(mode="json", by_alias=True, exclude_defaults=not include_type)

    def as_id(self) -> Self:
        return self



from datetime import datetime

from pydantic import Field, JsonValue

from cognite_toolkit._cdf_tk.constants import (
    INSTANCE_EXTERNAL_ID_PATTERN,
    SPACE_FORMAT_PATTERN,
)

from .base import ToolkitResource
from .view_field_definitions import DirectRelationReference, ViewReference


class CogniteFileYAML(ToolkitResource):
    space: str = Field(description="The space where the file is located.", max_length=43, pattern=SPACE_FORMAT_PATTERN)
    external_id: str = Field(
        description="External-id of the file.", max_length=255, pattern=INSTANCE_EXTERNAL_ID_PATTERN
    )
    name: str | None = Field(default=None, description="name of the file.")
    description: str | None = Field(default=None, description="Description of the file.")
    tags: list[str] | None = Field(default=None, description="Text based labels for generic use.", max_length=1000)
    aliases: list[str] | None = Field(default=None, description="Alternative names for the file.")
    source_id: str | None = Field(default=None, description="Identifier from the source system.")
    source_context: str | None = Field(default=None, description="Context of the source system.")
    source: DirectRelationReference | None = Field(default=None, description="Direct relation to the source system.")
    source_created_time: datetime | None = Field(
        default=None, description="When the file was created in the source system."
    )
    source_updated_time: datetime | None = Field(
        default=None, description="When the file was last updated in the source system."
    )
    source_created_user: str | None = Field(
        default=None, description="User identifier from the source system who created the file."
    )
    source_updated_user: str | None = Field(
        default=None, description="User identifier from the source system who last updated the file."
    )
    assets: list[DirectRelationReference] | None = Field(
        default=None, description="A list of assets this file is related to."
    )
    mime_type: str | None = Field(default=None, description="MIME type of the file.")
    directory: str | None = Field(default=None, description="Directory path for the file.")
    category: DirectRelationReference | None = Field(
        default=None, description="Direct relation to the category of the file."
    )
    existing_version: int | None = Field(default=None, description="Existing version of the file.")
    type: DirectRelationReference | None = Field(default=None, description="Direct relation to the type of the file.")
    node_source: ViewReference | None = Field(default=None, description="The source view for this file.")
    extra_properties: dict[str, JsonValue] | None = Field(
        default=None, description="Additional custom properties for the file."
    )



from collections import defaultdict
from collections.abc import Callable, Iterable, Sequence
from typing import Any

from cognite_toolkit._cdf_tk.builders import Builder
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase
from cognite_toolkit._cdf_tk.cruds import RawDatabaseCRUD, RawTableCRUD, ResourceCRUD
from cognite_toolkit._cdf_tk.data_classes import (
    BuildDestinationFile,
    BuildSourceFile,
    ModuleLocation,
    SourceLocation,
    SourceLocationEager,
)
from cognite_toolkit._cdf_tk.tk_warnings import ToolkitWarning
from cognite_toolkit._cdf_tk.utils import calculate_hash
from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump


class RawBuilder(Builder):
    _resource_folder = RawDatabaseCRUD.folder_name

    def build(
        self, source_files: list[BuildSourceFile], module: ModuleLocation, console: Callable[[str], None] | None = None
    ) -> Iterable[BuildDestinationFile | Sequence[ToolkitWarning]]:
        for source_file in source_files:
            loaded = source_file.loaded
            if loaded is None:
                continue
            loaded_list = loaded if isinstance(loaded, list) else [loaded]
            seen_databases: set[tuple] = set()
            entry_by_loader: dict[type[ResourceCRUD], list[dict[str, Any]]] = defaultdict(list)
            has_split_table_and_database = False

            for item in loaded_list:
                try:
                    table_id = RawTableCRUD.get_id(item)
                except KeyError:
                    seen_databases.add(tuple(item.items()))
                    entry_by_loader[RawDatabaseCRUD].append(item)
                else:
                    entry_by_loader[RawTableCRUD].append(item)
                    db_item = RawDatabaseCRUD.dump_id(RawDatabase(table_id.db_name))
                    hashable_db_item = tuple(db_item.items())
                    if hashable_db_item not in seen_databases:
                        seen_databases.add(hashable_db_item)
                        entry_by_loader[RawDatabaseCRUD].append(db_item)
                        has_split_table_and_database = True

            for loader, entries in entry_by_loader.items():
                if not entries:
                    continue
                destination_path = self._create_destination_path(source_file.source.path, loader.kind)

                if loader is RawDatabaseCRUD and has_split_table_and_database:
                    # We have inferred the database from a Table file, so we need to recalculate the hash
                    # in case we also inferred the database from another Table file
                    new_hash = calculate_hash(
                        yaml_safe_dump(sorted(entries, key=lambda entry: entry["dbName"])),
                        shorten=True,
                    )
                    source: SourceLocation = SourceLocationEager(path=source_file.source.path, _hash=new_hash)
                else:
                    source = source_file.source

                yield BuildDestinationFile(
                    path=destination_path,
                    loaded=entries,
                    loader=loader,
                    source=source,
                    extra_sources=None,
                )



from typing import Literal

from pydantic import Field, JsonValue

from cognite_toolkit._cdf_tk.resource_classes.base import ToolkitResource


class RobotMapYAML(ToolkitResource):
    name: str = Field(description="Map name.")
    external_id: str = Field(description="Map external id.", max_length=256)
    map_type: Literal["WAYPOINTMAP", "THREEDMODEL", "TWODMAP", "POINTCLOUD"] = Field(description="Map type.")
    description: str | None = Field(default=None, description="Description of Map. Textual description of the Map.")
    frame_external_id: str | None = Field(default=None, description="External id of the map's reference frame.")
    data: dict[str, JsonValue] | None = Field(default=None, description="Map-specific data.")
    location_external_id: str | None = Field(default=None, description="External id of the location.")
    scale: float | None = Field(default=None, description="Uniform scaling factor.", ge=0.0, le=1.0)



from pathlib import Path

from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses
from cognite_toolkit._cdf_tk.utils.fileio import COMPRESSION_BY_SUFFIX

from ._annotations import FileAnnotationIO
from ._applications import CanvasIO, ChartIO
from ._asset_centric import (
    AssetIO,
    BaseAssetCentricIO,
    EventIO,
    FileMetadataIO,
    HierarchyIO,
    TimeSeriesIO,
)
from ._base import (
    ConfigurableStorageIO,
    Page,
    StorageIO,
    StorageIOConfig,
    T_Selector,
    TableStorageIO,
    UploadableStorageIO,
    UploadItem,
)
from ._data_classes import InstanceIdCSVList, InstanceIdRow, ModelList
from ._datapoints import DatapointsIO
from ._instances import InstanceIO
from ._raw import RawIO
from .selectors._base import DataSelector

# MyPy fails to recognize that get_concrete_subclasses always returns a list of concrete subclasses.
STORAGE_IO_CLASSES = get_concrete_subclasses(StorageIO)  # type: ignore[type-abstract]
UPLOAD_IO_CLASSES = get_concrete_subclasses(UploadableStorageIO)  # type: ignore[type-abstract]


def get_upload_io(selector_cls: type[DataSelector], kind: str | Path) -> type[UploadableStorageIO]:
    """Get the appropriate UploadableStorageIO class based on the type of the provided selector."""
    for cls in UPLOAD_IO_CLASSES:
        if issubclass(selector_cls, cls.BASE_SELECTOR) and are_same_kind(cls.KIND, kind):
            return cls
    raise ValueError(f"No UploadableStorageIO found for selector of type {selector_cls.__name__}")


def are_same_kind(kind: str, kind_or_path: str | Path, /) -> bool:
    """Check if two kinds are the same, ignoring case and compression suffixes."""
    if not isinstance(kind_or_path, Path):
        return kind.casefold() == kind_or_path.casefold()
    stem = kind_or_path.stem
    if kind_or_path.suffix in COMPRESSION_BY_SUFFIX:
        stem = Path(stem).stem
    return stem.lower().endswith(kind.casefold())


__all__ = [
    "AssetIO",
    "BaseAssetCentricIO",
    "CanvasIO",
    "ChartIO",
    "ConfigurableStorageIO",
    "DatapointsIO",
    "EventIO",
    "FileAnnotationIO",
    "FileMetadataIO",
    "HierarchyIO",
    "InstanceIO",
    "InstanceIdCSVList",
    "InstanceIdRow",
    "ModelList",
    "Page",
    "RawIO",
    "StorageIO",
    "StorageIOConfig",
    "T_Selector",
    "TableStorageIO",
    "TimeSeriesIO",
    "UploadItem",
    "UploadableStorageIO",
    "are_same_kind",
    "get_upload_io",
]



"""Toolkit resource classes are pydantic models that represent the YAML file format the Cognite Toolkit uses.

This is means that we have three set of resource classes we use in Toolkit:

1. Toolkit resource classes (this module): Represent the YAML configuration format the Toolkit uses.
2. Write resource classes (from cognite-sdk): Represent the write/request format the Cognite resources.
3. Read resource classes (from cognite-sdk): Represent the read/response format the Cognite resources.
"""

from .agent import AgentYAML
from .asset import AssetYAML
from .base import BaseModelResource, ToolkitResource
from .cognitefile import CogniteFileYAML
from .containers import ContainerYAML
from .data_model import DataModelYAML
from .datapoint_subscription import DatapointSubscriptionYAML
from .dataset import DataSetYAML
from .event import EventYAML
from .extraction_pipeline import ExtractionPipelineYAML
from .extraction_pipeline_config import ExtractionPipelineConfigYAML
from .filemetadata import FileMetadataYAML
from .function_schedule import FunctionScheduleYAML
from .functions import FunctionsYAML
from .graphql_model import GraphQLDataModelYAML
from .groups import GroupYAML
from .hosted_extractor_destination import HostedExtractorDestinationYAML
from .hosted_extractor_job import HostedExtractorJobYAML
from .hosted_extractor_mapping import HostedExtractorMappingYAML
from .hosted_extractor_source import HostedExtractorSourceYAML
from .infield_location_config import InfieldLocationConfigYAML
from .infield_v1 import InfieldV1YAML
from .instance import EdgeYAML, NodeYAML
from .labels import LabelsYAML
from .location import LocationYAML
from .migration import ResourceViewMappingYAML
from .raw_database_table import DatabaseYAML, TableYAML
from .relationship import RelationshipYAML
from .robotics import RobotCapabilityYAML, RobotDataPostProcessingYAML, RobotFrameYAML, RobotLocationYAML, RobotMapYAML
from .search_config import SearchConfigYAML
from .securitycategories import SecurityCategoriesYAML
from .sequence import SequenceRowYAML, SequenceYAML
from .space import SpaceYAML
from .streamlit_ import StreamlitYAML
from .threedmodels import ThreeDModelYAML
from .timeseries import TimeSeriesYAML
from .transformation_notification import TransformationNotificationYAML
from .transformation_schedule import TransformationScheduleYAML
from .transformations import TransformationYAML
from .views import ViewYAML
from .workflow import WorkflowYAML
from .workflow_trigger import WorkflowTriggerYAML
from .workflow_version import WorkflowVersionYAML

__all__ = [
    "AgentYAML",
    "AssetYAML",
    "BaseModelResource",
    "CogniteFileYAML",
    "ContainerYAML",
    "DataModelYAML",
    "DataModelYAML",
    "DataSetYAML",
    "DatabaseYAML",
    "DatapointSubscriptionYAML",
    "EdgeYAML",
    "EventYAML",
    "ExtractionPipelineConfigYAML",
    "ExtractionPipelineYAML",
    "FileMetadataYAML",
    "FunctionScheduleYAML",
    "FunctionsYAML",
    "GraphQLDataModelYAML",
    "GroupYAML",
    "HostedExtractorDestinationYAML",
    "HostedExtractorJobYAML",
    "HostedExtractorMappingYAML",
    "HostedExtractorSourceYAML",
    "InfieldLocationConfigYAML",
    "InfieldV1YAML",
    "LabelsYAML",
    "LocationYAML",
    "NodeYAML",
    "RelationshipYAML",
    "ResourceViewMappingYAML",
    "RobotCapabilityYAML",
    "RobotDataPostProcessingYAML",
    "RobotFrameYAML",
    "RobotLocationYAML",
    "RobotMapYAML",
    "SearchConfigYAML",
    "SecurityCategoriesYAML",
    "SequenceRowYAML",
    "SequenceYAML",
    "SpaceYAML",
    "StreamlitYAML",
    "TableYAML",
    "ThreeDModelYAML",
    "TimeSeriesYAML",
    "ToolkitResource",
    "TransformationNotificationYAML",
    "TransformationScheduleYAML",
    "TransformationYAML",
    "ViewYAML",
    "WorkflowTriggerYAML",
    "WorkflowVersionYAML",
    "WorkflowYAML",
]



from .capability import RobotCapabilityYAML
from .data_postprocessing import RobotDataPostProcessingYAML
from .frame import RobotFrameYAML
from .location import RobotLocationYAML
from .map import RobotMapYAML

__all__ = [
    "RobotCapabilityYAML",
    "RobotDataPostProcessingYAML",
    "RobotFrameYAML",
    "RobotLocationYAML",
    "RobotMapYAML",
]



import tempfile
from pathlib import Path
from zipfile import ZipFile

import pytest

from cognite_toolkit._cdf_tk.utils.file import create_temporary_zip, read_yaml_content, sanitize_filename


class TestCreateTemporaryZip:
    def test_create_temporary_zip(self) -> None:
        with tempfile.TemporaryDirectory() as test_dir:
            test_dir_path = Path(test_dir)
            # Create test directory structure
            subdir = test_dir_path / "subdir"
            subdir.mkdir()
            # Create some test files
            (test_dir_path / "file1.txt").write_text("file1 content")
            (test_dir_path / "file2.txt").write_text("file2 content")
            (subdir / "file3.txt").write_text("file3 content")

            # Use the context manager to create a zip
            original_dir = Path.cwd()
            with create_temporary_zip(test_dir_path, "test.zip") as zip_path:
                # Verify the zip file exists
                assert zip_path.exists()
                assert zip_path.name == "test.zip"

                # Verify the contents of the zip file
                with ZipFile(zip_path, "r") as zip_file:
                    zip_contents = zip_file.namelist()
                    expected_files = {"subdir/", "subdir/file3.txt", "./", "file2.txt", "file1.txt"}
                    assert set(zip_contents) == expected_files

            # Verify we're back in the original directory during zip creation
            assert Path.cwd() == original_dir


class TestSanitizeFilename:
    @pytest.mark.parametrize(
        "filename, expected",
        [
            ("valid_filename.yaml", "valid_filename.yaml"),
            ("another-valid_filename123.json", "another-valid_filename123.json"),
            ("invalid/filename.yaml", "invalid_filename.yaml"),
            ("invalid\\filename.json", "invalid_filename.json"),
            ("invalid:filename.yaml", "invalid_filename.yaml"),
            ("invalid*filename.json", "invalid_filename.json"),
            ("invalid?filename.yaml", "invalid_filename.yaml"),
            ('invalid"filename.json', "invalid_filename.json"),
            ("invalid<filename.yaml", "invalid_filename.yaml"),
            ("invalid>filename.json", "invalid_filename.json"),
            ("invalid|filename.yaml", "invalid_filename.yaml"),
            ("inva|lid:fi*le?name<.json", "inva_lid_fi_le_name_.json"),
        ],
    )
    def test_sanitize_filename(self, filename: str, expected: str) -> None:
        assert sanitize_filename(filename) == expected


class TestReadYamlContent:
    def test_keyvault_tag_preserved_scalar(self) -> None:
        """Test that !keyvault tag is preserved along with the value."""
        yaml_content = "password: !keyvault value-secret-name"
        result = read_yaml_content(yaml_content)
        assert result == {"password": "!keyvault value-secret-name"}

    def test_keyvault_tag_preserved_in_nested_structure(self) -> None:
        """Test that !keyvault tag is preserved in nested YAML structures."""
        yaml_content = """azure-keyvault:
  authentication-method: client-secret
  password: !keyvault value-secret-name
databases:
-   connection-string: !keyvault db-connection-secret
    name: my_db"""
        result = read_yaml_content(yaml_content)
        assert result["azure-keyvault"]["password"] == "!keyvault value-secret-name"
        assert result["databases"][0]["connection-string"] == "!keyvault db-connection-secret"
        assert result["databases"][0]["name"] == "my_db"

    def test_keyvault_tag_preserved_multiple_values(self) -> None:
        """Test that multiple !keyvault tags are preserved."""
        yaml_content = """secret1: !keyvault secret-name-1
secret2: !keyvault secret-name-2
normal_value: regular-string"""
        result = read_yaml_content(yaml_content)
        assert result["secret1"] == "!keyvault secret-name-1"
        assert result["secret2"] == "!keyvault secret-name-2"
        assert result["normal_value"] == "regular-string"



import queue
import sys
import threading
import time
import traceback
import typing
from collections.abc import Callable, Iterable, Sized
from typing import Any, Generic, TypeVar

from rich.console import Console
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    Task,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)

from cognite_toolkit._cdf_tk.exceptions import ToolkitRuntimeError

T_Download = TypeVar("T_Download", bound=Sized)
T_Processed = TypeVar("T_Processed", bound=Sized)
T_Item = TypeVar("T_Item", bound=Sized)

# Sentinels for signaling finish
PROCESS_FINISH_SENTINEL = object()
WRITE_FINISH_SENTINEL = object()


class ItemCountColumn(ProgressColumn):
    def render(self, task: Task) -> str:
        return f"[green]{int(task.fields.get('item_count', 0)):,} items"


class ProducerWorkerExecutor(Generic[T_Download, T_Processed]):
    """This class manages the execution of a producer-worker pattern with multiple threads.

    It downloads data from an iterable, processes it, and writes the processed data.

    Args:
        download_iterable (Iterable[T_Download]): An iterable that yields chunks of items.
        process (Callable[[T_Download], T_Processed]): A function that processes the chunks of items.
        write (Callable[[T_Processed], None]): A function that writes the processed items. This is typically to
            a file, CDF, or another storage format.
        iteration_count (int): The total number of iterations expected. Note this is not the total number of items,
            but rather the number of chunks of items that will be processed. Set to `None` if the total number of
            iterations is unknown.
        max_queue_size (int): The maximum size of the queues used for processing and writing. Note that the queue
            holds chunks of items, not individual items, so this should be set based on the expected size of the
            chunks being processed.
        download_description (str): A description of the download process, used for progress tracking.
        process_description (str): A description of the processing step, used for progress tracking.
        write_description (str): A description of the writing step, used for progress tracking.
        console (Console | None): An optional Rich Console instance for outputting progress and error messages.

    Examples:
        >>> from cognite_toolkit._cdf_tk.utils.producer_worker import ProducerWorkerExecutor
        >>> from typing import List
        >>> def download_data() -> Iterable[list[int]]:
        ...     yield [1, 2, 3, 4, 5]
        ...     yield [6, 7, 8, 9, 10]
        >>> def process_data(data: list[int]) -> list[int]:
        ...     return [x * 2 for x in data]
        >>> def write_data(data: list[int]) -> None:
        ...     print(f"Writing data: {data}")
        >>> executor = ProducerWorkerExecutor(
        ...     download_iterable=download_data(),
        ...     process=process_data,
        ...     write=write_data,
        ...     iteration_count=2,
        ...     max_queue_size=1,
        ... )
        >>> executor.run()

    """

    def __init__(
        self,
        download_iterable: Iterable[T_Download],
        process: Callable[[T_Download], T_Processed],
        write: Callable[[T_Processed], None],
        iteration_count: int | None,
        max_queue_size: int,
        download_description: str = "Producing",
        process_description: str = "Processing",
        write_description: str = "Writing",
        console: Console | None = None,
        verbose: bool = False,
    ) -> None:
        self._download_iterable = download_iterable
        self._process = process
        self._write = write
        self.iteration_count = iteration_count
        self.download_description = download_description
        self.process_description = process_description
        self.write_description = write_description
        self.console = console or Console()
        self._stop_event = threading.Event()
        self._error_event = threading.Event()
        # Queues for managing the flow of data between threads
        # Download -> [process_queue] -> Process -> [write_queue] -> Write
        self.process_queue: queue.Queue[T_Download] = queue.Queue(maxsize=max_queue_size)
        self.write_queue: queue.Queue[T_Processed] = queue.Queue(maxsize=max_queue_size)
        self.total_items = 0
        self.error_message = ""
        self.error_traceback = ""
        self.verbose = verbose

    @property
    def error_occurred(self) -> bool:
        """Indicates if an error occurred during execution."""
        return self._error_event.is_set()

    @property
    def stopped_by_user(self) -> bool:
        """Indicates if the execution was stopped by the user."""
        return self._stop_event.is_set()

    def _get_progress_columns(self) -> list[ProgressColumn]:
        """Helper to set up the progress bar and tasks based on iteration_count."""
        if self.iteration_count is None:
            return [
                SpinnerColumn(),
                TextColumn("[bold blue]{task.description}"),
                ItemCountColumn(),
                TimeElapsedColumn(),
            ]
        else:
            return [
                TextColumn("[bold blue]{task.description}"),
                BarColumn(),
                TaskProgressColumn(),
                TimeRemainingColumn(),
            ]

    def _user_input_listener(self, producer_thread: threading.Thread) -> None:
        while not self._error_event.is_set() and not self._stop_event.is_set():
            key = getch(timeout=0.1)
            if key is None and not producer_thread.is_alive():
                break
            elif key is None:
                continue
            elif key.casefold() == "q":
                self._stop_event.set()
                self.console.print(
                    f"[yellow]Execution stopped by user. Finishing: {self.write_description}...[/yellow]"
                )
                break

    def run(self) -> None:
        self.console.print(f"[blue]Starting {self.download_description} (Press 'q' to stop)...[/blue]")
        columns = self._get_progress_columns()
        with Progress(*columns, console=self.console) as progress:
            task_args: dict[str, Any] = (
                {"item_count": 0, "total": None} if self.iteration_count is None else {"total": self.iteration_count}
            )
            download_task = progress.add_task(self.download_description, **task_args)
            process_task = progress.add_task(self.process_description, **task_args)
            write_task = progress.add_task(self.write_description, **task_args)

            download_thread = threading.Thread(target=self._download_worker, args=(progress, download_task))
            process_thread = threading.Thread(target=self._process_worker, args=(progress, process_task))
            write_thread = threading.Thread(target=self._write_worker, args=(progress, write_task))

            download_thread.start()
            process_thread.start()
            write_thread.start()

            input_thread = threading.Thread(target=self._user_input_listener, args=(download_thread,))
            input_thread.start()

            for t in [download_thread, process_thread, write_thread]:
                try:
                    t.join()
                except KeyboardInterrupt:
                    self.console.print("[red]Execution interrupted by user.[/red]")
                    self._stop_event.set()
                    break

            # After a possible interrupt, we must wait for all threads to finish their
            # graceful shutdown. This is important to prevent data loss.
            for t in [download_thread, process_thread, write_thread, input_thread]:
                if t.is_alive():
                    t.join()

    def raise_on_error(self) -> None:
        """Raises an exception if an error occurred during execution."""
        if self._error_event.is_set():
            if self.verbose and self.error_traceback:
                self.print_traceback()
            raise ToolkitRuntimeError(f"An error occurred during execution: {self.error_message}")
        if self._stop_event.is_set():
            raise ToolkitRuntimeError("Execution was stopped by the user.")

    def print_traceback(self) -> None:
        """Prints the traceback if an error occurred during execution."""
        if self.error_traceback:
            self.console.print(
                Panel(
                    self.error_traceback,
                    title="Traceback",
                    expand=False,
                    border_style="red",
                )
            )

    def _download_worker(self, progress: Progress, download_task: TaskID) -> None:
        """Worker thread for downloading data."""
        try:
            iterator = iter(self._download_iterable)
        except TypeError as e:
            self._error_event.set()
            self.error_message = str(e)
            self.console.print(f"[red]Error[/red] occurred while {self.download_description}: {self.error_message}")
            return
        item_count = 0
        while not self._error_event.is_set():
            try:
                if self._stop_event.is_set():
                    break
                items = next(iterator)
                self.total_items += len(items)
                if self._put_with_error_check(items, self.process_queue):
                    item_count += len(items)
                    progress.update(download_task, advance=1, item_count=item_count)
                    continue
                break  # Exit if error event was set while waiting to put
            except StopIteration:
                break
            except Exception as e:
                self._error_event.set()
                self.error_message = f"{type(e).__name__} {e!s}"
                self.error_traceback = traceback.format_exc()
                self.console.print(f"[red]Error[/red] occurred while {self.download_description}: {self.error_message}")
                break
        self._put_with_error_check(PROCESS_FINISH_SENTINEL, self.process_queue)  # type: ignore[misc]

    def _put_with_error_check(self, items: T_Item, target_queue: queue.Queue[T_Item]) -> bool:
        """Helper to put items into a queue with error checking."""
        while not self._error_event.is_set():
            try:
                target_queue.put(items, timeout=0.5)
                return True
            except queue.Full:
                continue
        return False

    def _process_worker(self, progress: Progress, process_task: TaskID) -> None:
        """Worker thread for processing data."""
        item_count = 0
        while not self._error_event.is_set():
            try:
                items = self.process_queue.get(timeout=0.5)
                if items is PROCESS_FINISH_SENTINEL:
                    # Signal writer to finish
                    self._put_with_error_check(WRITE_FINISH_SENTINEL, self.write_queue)  # type: ignore[misc]
                    self.process_queue.task_done()
                    break
                processed_items = self._process(items)
                if self._put_with_error_check(processed_items, self.write_queue):
                    item_count += len(items)
                    progress.update(process_task, advance=1, item_count=item_count)
                    self.process_queue.task_done()
                    continue
                else:
                    self.process_queue.task_done()
                    break  # Exit if error event was set while waiting to put
            except queue.Empty:
                continue
            except Exception as e:
                self._error_event.set()
                self.error_message = f"{type(e).__name__} {e!s}"
                self.error_traceback = traceback.format_exc()
                self.console.print(f"[red]Error[/red] occurred while {self.process_description}: {self.error_message}")
                break

    def _write_worker(self, progress: Progress, write_task: TaskID) -> None:
        """Worker thread for writing data to file."""
        item_count = 0
        while not self._error_event.is_set():
            try:
                items = self.write_queue.get(timeout=0.5)
                if items is WRITE_FINISH_SENTINEL:
                    self.write_queue.task_done()
                    break
                self._write(items)
                item_count += len(items)
                progress.update(write_task, advance=1, item_count=item_count)
                self.write_queue.task_done()
            except queue.Empty:
                continue
            except Exception as e:
                self._error_event.set()
                self.error_message = f"{type(e).__name__} {e!s}"
                self.error_traceback = traceback.format_exc()
                self.console.print(f"[red]Error[/red] occurred while {self.write_description}: {self.error_message}")
                break


# MyPy fails as the imports are os specific
# thus we disable type checking for this function
@typing.no_type_check
def getch(timeout: float) -> str | None:
    """Get a single character from standard input. Does not echo to the screen."""
    if not sys.stdin.isatty():
        return None
    try:
        # Windows
        import msvcrt

        end_time = time.time() + timeout
        while time.time() < end_time:
            if msvcrt.kbhit():
                return msvcrt.getwch()
            time.sleep(0.01)
        return None
    except ImportError:
        # Unix
        import select
        import termios
        import tty

        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            rlist, _, _ = select.select([fd], [], [], timeout)
            if rlist:
                ch = sys.stdin.read(1)
                return ch
            else:
                return None
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)



from unittest.mock import MagicMock
from warnings import catch_warnings

import pytest
from cognite.client.data_classes.data_modeling import NodeId, ViewId
from cognite.client.data_classes.data_modeling.instances import Node, Properties

from cognite_toolkit._cdf_tk.client.api.extended_data_modeling import ExtendedInstancesAPI
from cognite_toolkit._cdf_tk.client.api.migration import ResourceViewMappingAPI
from cognite_toolkit._cdf_tk.client.data_classes.migration import ResourceViewMapping, ResourceViewMappingApply
from cognite_toolkit._cdf_tk.constants import COGNITE_MIGRATION_SPACE


@pytest.fixture()
def instance_api() -> MagicMock:
    return MagicMock(spec=ExtendedInstancesAPI)


class TestViewSource:
    def test_upsert(self, instance_api: MagicMock) -> None:
        view_source_api = ResourceViewMappingAPI(instance_api=instance_api)
        view_source = ResourceViewMappingApply(
            external_id="asset_mapping",
            resource_type="asset",
            view_id=ViewId("cdf_cdm", "CogniteAsset", "v1"),
            property_mapping={
                "name": "name",
                "description": "description",
            },
        )
        _ = view_source_api.upsert(view_source)

        assert instance_api.apply.called
        args, kwargs = instance_api.apply.call_args
        assert args[0] == view_source
        assert kwargs == {"skip_on_version_conflict": False, "replace": False}

    def test_retrieve_single_valid(self, instance_api: MagicMock) -> None:
        view_source_api = ResourceViewMappingAPI(instance_api=instance_api)
        mock_view_source = ResourceViewMapping(
            external_id="test_source",
            version=1,
            last_updated_time=1000,
            created_time=1000,
            resource_type="asset",
            view_id=ViewId("cdf_cdm", "CogniteAsset", "v1"),
            property_mapping={"name": "name"},
        )
        instance_api.retrieve_nodes.return_value = mock_view_source

        result = view_source_api.retrieve("test_source")
        assert result == mock_view_source
        instance_api.retrieve_nodes.assert_called_once()
        args, kwargs = instance_api.retrieve_nodes.call_args
        assert args[0] == mock_view_source.as_id()
        assert kwargs == {"node_cls": ResourceViewMapping}

    def test_retrieve_multiple_valid(self, instance_api: MagicMock) -> None:
        view_source_api = ResourceViewMappingAPI(instance_api=instance_api)
        mock_view_sources = [
            Node(
                space=COGNITE_MIGRATION_SPACE,
                external_id=f"test_source_{i}",
                version=1,
                last_updated_time=1000,
                created_time=1000,
                deleted_time=None,
                type=None,
                properties=Properties(
                    {
                        ResourceViewMappingApply.get_source(): {
                            "resourceType": "asset",
                            "viewId": ViewId("cdf_cdm", "CogniteAsset", "v1").dump(),
                            "propertyMapping": {"name": "name"},
                        }
                    }
                ),
            )
            for i in range(2)
        ]
        instance_api.retrieve.return_value.nodes = mock_view_sources

        result = view_source_api.retrieve(["test_source_0", "test_source_1"])
        assert len(result) == 2
        assert all(isinstance(r, ResourceViewMapping) for r in result)
        instance_api.retrieve.assert_called_once()

    def test_retrieve_single_invalid(self, instance_api: MagicMock) -> None:
        view_source_api = ResourceViewMappingAPI(instance_api=instance_api)
        instance_api.retrieve_nodes.side_effect = ValueError(
            "Invalid viewId format. Expected 'space', 'externalId', 'version'. Error: 'some_error'"
        )

        with pytest.raises(ValueError) as exc_info:
            _ = view_source_api.retrieve("invalid_source")

        assert (
            str(exc_info.value)
            == "Invalid viewId format. Expected 'space', 'externalId', 'version'. Error: 'some_error'"
        )

    def test_retrieve_multiple_some_invalid(self, instance_api: MagicMock) -> None:
        default_args = dict(
            space=COGNITE_MIGRATION_SPACE,
            version=1,
            last_updated_time=1000,
            created_time=1000,
            deleted_time=None,
            type=None,
        )
        view_source_api = ResourceViewMappingAPI(instance_api=instance_api)
        mock_view_sources = [
            Node(
                external_id="valid_source",
                **default_args,
                properties=Properties(
                    {
                        ResourceViewMappingApply.get_source(): {
                            "resourceType": "asset",
                            "viewId": ViewId("cdf_cdm", "CogniteAsset", "v1").dump(),
                            "propertyMapping": {"name": "name"},
                        }
                    }
                ),
            ),
            Node(
                external_id="invalid_source",
                **default_args,
                properties=Properties(
                    {
                        ResourceViewMappingApply.get_source(): {
                            "resourceType": "asset",
                            "viewId": "invalid_view_id_format",
                            "propertyMapping": {"Not a valid mapping": {"name": "name"}},
                        }
                    }
                ),
            ),
        ]
        instance_api.retrieve.return_value.nodes = mock_view_sources

        with catch_warnings(record=True) as record:
            result = view_source_api.retrieve(["valid_source", "invalid_source"])
        assert len(result) == 1
        first = result[0]
        assert isinstance(first, ResourceViewMapping)
        assert first.external_id == "valid_source"
        assert len(record) == 1

    def test_delete_single(self, instance_api: MagicMock) -> None:
        view_source_api = ResourceViewMappingAPI(instance_api=instance_api)
        instance_api.delete.return_value.nodes = ["deleted_node_id"]

        result = view_source_api.delete("test_source")
        assert result == "deleted_node_id"
        instance_api.delete.assert_called_once()
        args, _ = instance_api.delete.call_args
        assert args[0] == NodeId(COGNITE_MIGRATION_SPACE, "test_source")

    def test_delete_multiple(self, instance_api: MagicMock) -> None:
        view_source_api = ResourceViewMappingAPI(instance_api=instance_api)
        instance_api.delete.return_value.nodes = ["deleted_node_1", "deleted_node_2"]

        result = view_source_api.delete(["test_source_1", "test_source_2"])
        assert len(result) == 2
        instance_api.delete.assert_called_once()
        args, _ = instance_api.delete.call_args
        assert args[0] == [
            NodeId(COGNITE_MIGRATION_SPACE, "test_source_1"),
            NodeId(COGNITE_MIGRATION_SPACE, "test_source_2"),
        ]



from collections.abc import Hashable, Iterable, Sequence
from typing import Any

from cognite.client.data_classes.agents import Agent, AgentList, AgentUpsert, AgentUpsertList
from cognite.client.data_classes.capabilities import Capability
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import AgentYAML
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable


class AgentCRUD(ResourceCRUD[str, AgentUpsert, Agent, AgentUpsertList, AgentList]):
    folder_name = "agents"
    filename_pattern = r".*\.Agent$"  # Matches all yaml files whose stem ends with '.Agent'.
    resource_cls = Agent
    resource_write_cls = AgentUpsert
    list_cls = AgentList
    list_write_cls = AgentUpsertList
    kind = "Agent"
    yaml_cls = AgentYAML
    _doc_base_url = ""
    _doc_url = "https://api-docs.cognite.com/20230101-beta/tag/Agents/operation/main_ai_agents_post/"

    @classmethod
    def get_id(cls, item: AgentUpsert | Agent | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[AgentUpsert] | None, read_only: bool
    ) -> Capability | list[Capability]:
        return []

    def create(self, items: AgentUpsertList) -> AgentList:
        return self.client.agents.upsert(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> AgentList:
        return self.client.agents.retrieve(ids, ignore_unknown_ids=True)

    def update(self, items: AgentUpsertList) -> AgentList:
        return self.client.agents.upsert(items)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        try:
            self.client.agents.delete(ids)
        except CogniteAPIError:
            deleted = 0
            for id in ids:
                try:
                    self.client.agents.delete(id)
                    deleted += 1
                except CogniteAPIError:
                    # accepted because the resource may not exist
                    pass
            return deleted
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Agent]:
        return self.client.agents.list()

    def dump_resource(self, resource: Agent, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        # The atlas endpoints are not yet full implemented. There are properties being added and removed that are
        # not part of the official API. For example, as of 31.July 2025, `labels` is not part of the API, however,
        # this is necessary to ensure that the agents are shown as published in the UI, so we cannot ignore it.
        # The below logic ensures that we keep the unknown properties returned by the API, such that when we run
        # `cdf dump agents` we will not lose any properties that are not part of the official API.
        if (unknown_props := getattr(resource, "_unknown_properties", None)) and isinstance(unknown_props, dict):
            dumped.update(unknown_props)
        if local is None:
            return dumped
        if resource.instructions == "" and "instructions" not in local:
            # Instructions are optional, if not set the server set them to an empty string.
            # We remove them from the dumped resource to ensure it will be equal to the local resource.
            dumped.pop("instructions", None)
        for key in ["labels", "exampleQuestions"]:
            if key not in local and not dumped.get(key):
                # If the local resource does not have the key and the server set Agent has it set to an empty list,
                # we remove it from the dumped resource to ensure it will be equal to the local resource.
                dumped.pop(key, None)
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        """
        Compare two lists and return a mapping of local indices to CDF indices and a list of CDF indices that are not
        present in the local list.
        """
        if json_path == ("tools",):
            return diff_list_identifiable(
                local, cdf, get_identifier=lambda t: (t.get("name", ""), t.get("description", ""))
            )
        elif json_path == ("labels",):
            return diff_list_hashable(local, cdf)
        elif json_path == ("exampleQuestions",):
            return diff_list_identifiable(
                local, cdf, get_identifier=lambda q: q.get("question", "") if isinstance(q, dict) else str(q)
            )
        return super().diff_list(local, cdf, json_path)



from collections.abc import Sequence
from functools import partial
from pathlib import Path

from cognite.client.data_classes._base import T_CogniteResource
from pydantic import ValidationError
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.constants import DATA_MANIFEST_STEM, DATA_RESOURCE_DIR
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.storageio import (
    T_Selector,
    UploadableStorageIO,
    are_same_kind,
    get_upload_io,
)
from cognite_toolkit._cdf_tk.storageio._base import T_WriteCogniteResource, TableUploadableStorageIO, UploadItem
from cognite_toolkit._cdf_tk.storageio.selectors import Selector, SelectorAdapter
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, MediumSeverityWarning
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.file import read_yaml_file
from cognite_toolkit._cdf_tk.utils.fileio import TABLE_READ_CLS_BY_FORMAT, FileReader
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, ItemMessage, SuccessResponseItems
from cognite_toolkit._cdf_tk.utils.producer_worker import ProducerWorkerExecutor
from cognite_toolkit._cdf_tk.utils.progress_tracker import ProgressTracker
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal
from cognite_toolkit._cdf_tk.validation import humanize_validation_error

from ._base import ToolkitCommand
from .deploy import DeployCommand


class UploadCommand(ToolkitCommand):
    """Command for uploading data to CDF from files in a specified directory.

    This command reads files matching a specific pattern in the input directory,
    processes them using a StorageIO instance, and uploads the data to CDF.

    Attributes:
        _MAX_QUEUE_SIZE (int): The maximum size of the queue for processing items.
            Set to 80 to balance memory usage and processing speed.
    """

    _MAX_QUEUE_SIZE = 80
    _UPLOAD = "upload"

    def upload(
        self,
        input_dir: Path,
        client: ToolkitClient,
        deploy_resources: bool,
        dry_run: bool,
        verbose: bool,
        kind: str | None = None,
    ) -> None:
        """Uploads data from files in the specified input directory to CDF.

        Args:
            input_dir: The directory containing the files to upload. It is expected to
                have the structure defined below.
            client: An instance of ToolkitClient to interact with CDF.
            deploy_resources: If True, deploys resources from the 'resources' subdirectory.
            dry_run: If True, performs a dry run without actually uploading the data
                (or deploying resources).
            verbose: If True, prints detailed information about the upload process.
            kind: Optional; if provided, only data files of this kind will be processed.

        The expected structure of the input directory is as follows:
        ```
        input_dir/
        ├── resources/                # Optional, only if deploy_resources is True
        │   ├── raw/
        │   │   ├── table1.Table.yaml
        │   │   └── table2.Table.yaml
        │   └── ...
        ├── datafile1.kind.ndjson # Data file of a specific kind
        ├── datafile1.Metadata.yaml       # Metadata file for datafile1
        ├── datafile2.kind2.ndjson # Another data file of the same or different kind
        ├── datafile2.Metadata.yaml       # Metadata file for datafile2
        └── ...
        """
        console = Console()
        data_files_by_selector = self._find_data_files(input_dir, kind)

        self._deploy_resource_folder(input_dir / DATA_RESOURCE_DIR, deploy_resources, client, console, dry_run, verbose)

        self._upload_data(data_files_by_selector, client, dry_run, input_dir, console, verbose)

    def _find_data_files(
        self,
        input_dir: Path,
        kind: str | None = None,
    ) -> dict[Selector, list[Path]]:
        """Finds data files and their corresponding metadata files in the input directory."""
        manifest_file_endswith = f".{DATA_MANIFEST_STEM}.yaml"
        data_files_by_metadata: dict[Selector, list[Path]] = {}
        for metadata_file in input_dir.glob(f"*{manifest_file_endswith}"):
            data_file_prefix = metadata_file.name.removesuffix(manifest_file_endswith)
            data_files = [
                file
                for file in input_dir.glob(f"{data_file_prefix}*")
                if not file.name.endswith(manifest_file_endswith)
            ]
            if kind is not None and data_files:
                data_files = [data_file for data_file in data_files if are_same_kind(kind, data_file)]
                if not data_files:
                    continue
            if not data_files:
                self.warn(
                    MediumSeverityWarning(
                        f"Metadata file {metadata_file.as_posix()!r} has no corresponding data files, skipping.",
                    )
                )
                continue

            selector_dict = read_yaml_file(metadata_file, expected_output="dict")
            try:
                selector = SelectorAdapter.validate_python(selector_dict)
            except ValidationError as e:
                errors = humanize_validation_error(e)
                self.warn(
                    ResourceFormatWarning(
                        metadata_file, tuple(errors), text="Invalid selector in metadata file, skipping."
                    )
                )
                continue
            data_files_by_metadata[selector] = data_files
        return data_files_by_metadata

    def _deploy_resource_folder(
        self,
        resource_dir: Path,
        deploy_resources: bool,
        client: ToolkitClient,
        console: Console,
        dry_run: bool,
        verbose: bool,
    ) -> None:
        """Deploy resources from the specified resource directory if it exists and deployment is enabled."""
        if deploy_resources and resource_dir.exists():
            deploy_command = DeployCommand()
            deploy_results = deploy_command.deploy_all_resources(
                client=client,
                build_dir=resource_dir,
                env_vars=EnvironmentVariables.create_from_environment(),
                dry_run=dry_run,
                verbose=verbose,
            )
            self.warning_list.extend(deploy_command.warning_list)
            if deploy_results.has_counts:
                console.print(deploy_results.counts_table())
        elif deploy_resources:
            self.warn(
                MediumSeverityWarning(
                    f"Resource directory {resource_dir.as_posix()!r} does not exist, skipping resource deployment."
                )
            )

    def _upload_data(
        self,
        data_files_by_selector: dict[Selector, list[Path]],
        client: ToolkitClient,
        dry_run: bool,
        input_dir: Path,
        console: Console,
        verbose: bool,
    ) -> None:
        total_file_count = sum(len(files) for files in data_files_by_selector.values())
        if verbose:
            input_dir_display = self._path_as_display_name(input_dir)
            console.print(f"Found {total_file_count} files to upload in {input_dir_display.as_posix()!r}.")
        action = "Would upload" if dry_run else "Uploading"
        with HTTPClient(config=client.config) as upload_client:
            file_count = 1
            for selector, datafiles in data_files_by_selector.items():
                io = self._create_selected_io(selector, datafiles[0], client)
                if io is None:
                    continue
                for data_file in datafiles:
                    file_display = self._path_as_display_name(data_file)
                    if verbose:
                        console.print(f"{action} {selector.display_name} from {file_display.as_posix()!r}")
                    reader = FileReader.from_filepath(data_file)
                    is_table = reader.format in TABLE_READ_CLS_BY_FORMAT
                    if is_table and not isinstance(io, TableUploadableStorageIO):
                        raise ToolkitValueError(f"{selector.display_name} does not support {reader.format!r} files.")
                    tracker = ProgressTracker[str]([self._UPLOAD])
                    executor = ProducerWorkerExecutor[list[tuple[str, dict[str, JsonVal]]], Sequence[UploadItem]](
                        download_iterable=io.read_chunks(reader),
                        process=partial(io.rows_to_data, selector=selector)
                        if is_table and isinstance(io, TableUploadableStorageIO)
                        else io.json_chunk_to_data,
                        write=partial(
                            self._upload_items,
                            upload_client=upload_client,
                            io=io,
                            dry_run=dry_run,
                            selector=selector,
                            tracker=tracker,
                            console=console,
                        ),
                        iteration_count=None,
                        max_queue_size=self._MAX_QUEUE_SIZE,
                        download_description=f"Reading {file_count:,}/{total_file_count + 1:,}: {file_display.as_posix()!s}",
                        process_description="Processing",
                        write_description=f"{action} {selector.display_name!r}",
                        console=console,
                    )
                    executor.run()
                    file_count += 1
                    executor.raise_on_error()
                    final_action = "Uploaded" if not dry_run else "Would upload"
                    suffix = " successfully" if not dry_run else ""
                    results = tracker.aggregate()
                    success = results.get((self._UPLOAD, "success"), 0)
                    failed = results.get((self._UPLOAD, "failed"), 0)
                    if failed > 0:
                        suffix += f", {failed:,} failed"
                    console.print(
                        f"{final_action} {success:,} {selector.display_name} from {file_display.as_posix()!r}{suffix}."
                    )

    @staticmethod
    def _path_as_display_name(input_path: Path, cwd: Path = Path.cwd()) -> Path:
        display_name = input_path
        if input_path.is_relative_to(cwd):
            display_name = input_path.relative_to(cwd)
        return display_name

    def _create_selected_io(
        self, selector: Selector, data_file: Path, client: ToolkitClient
    ) -> UploadableStorageIO | None:
        try:
            io_cls = get_upload_io(type(selector), kind=data_file)
        except ValueError as e:
            self.warn(HighSeverityWarning(f"Could not find StorageIO for selector {selector}: {e}"))
            return None
        return io_cls(client)

    @classmethod
    def _upload_items(
        cls,
        data_chunk: Sequence[UploadItem],
        upload_client: HTTPClient,
        io: UploadableStorageIO[T_Selector, T_CogniteResource, T_WriteCogniteResource],
        selector: T_Selector,
        dry_run: bool,
        tracker: ProgressTracker[str],
        console: Console,
    ) -> None:
        if dry_run:
            for item in data_chunk:
                tracker.set_progress(item.source_id, cls._UPLOAD, "success")
            return
        results = io.upload_items(data_chunk, upload_client, selector)
        for message in results:
            if isinstance(message, SuccessResponseItems):
                for id_ in message.ids:
                    tracker.set_progress(id_, step=cls._UPLOAD, status="success")
            elif isinstance(message, ItemMessage):
                for id_ in message.ids:
                    tracker.set_progress(id_, step=cls._UPLOAD, status="failed")
            else:
                console.log(f"[red]Unexpected result from upload: {str(message)!r}[/red]")



from functools import lru_cache

from cognite.client.data_classes.data_modeling import ViewId

from cognite_toolkit._cdf_tk.client.data_classes.migration import ResourceViewMappingApply

ASSET_ID = "cdf_asset_mapping"
EVENT_ID = "cdf_event_mapping"
TIME_SERIES_ID = "cdf_time_series_mapping"
TIME_SERIES_EXTRACTOR_ID = "cdf_time_series_extractor_mapping"
FILE_METADATA_ID = "cdf_file_metadata_mapping"
FILE_METADATA_EXTRACTOR_ID = "cdf_file_metadata_extractor_mapping"
FILE_ANNOTATIONS_ID = "cdf_file_annotations_mapping"
ASSET_ANNOTATIONS_ID = "cdf_asset_annotations_mapping"


@lru_cache(maxsize=1)
def create_default_mappings() -> list[ResourceViewMappingApply]:
    """Return the default mappings for migration."""
    ts_property_mapping = {
        "name": "name",
        "description": "description",
        "isStep": "isStep",
        "isString": "type",
        "unit": "sourceUnit",
        "unitExternalId": "unit",
        "assetId": "assets",
    }
    file_property_mapping = {
        "name": "name",
        "source": "source",
        "labels": "tags",
        "mimeType": "mimeType",
        "directory": "directory",
        "assetIds": "assets",
        "sourceCreatedTime": "sourceCreatedTime",
        "sourceModifiedTime": "sourceUpdatedTime",
    }

    return [
        ResourceViewMappingApply(
            external_id=ASSET_ID,
            resource_type="asset",
            view_id=ViewId("cdf_cdm", "CogniteAsset", "v1"),
            property_mapping={
                "name": "name",
                "description": "description",
                "source": "source",
                "labels": "tags",
                "parentId": "parent",
            },
        ),
        ResourceViewMappingApply(
            external_id=EVENT_ID,
            resource_type="event",
            view_id=ViewId("cdf_cdm", "CogniteActivity", "v1"),
            property_mapping={
                "startTime": "startTime",
                "endTime": "endTime",
                "description": "description",
                "source": "source",
                "labels": "tags",
                "assetIds": "assets",
            },
        ),
        ResourceViewMappingApply(
            external_id=TIME_SERIES_ID,
            resource_type="timeseries",
            view_id=ViewId("cdf_cdm", "CogniteTimeSeries", "v1"),
            property_mapping=ts_property_mapping,
        ),
        ResourceViewMappingApply(
            external_id=TIME_SERIES_EXTRACTOR_ID,
            resource_type="timeseries",
            view_id=ViewId("cdf_extraction_extensions", "CogniteExtractorTimeSeries", "v1"),
            property_mapping={**ts_property_mapping, "metadata": "extractedData"},
        ),
        ResourceViewMappingApply(
            external_id=FILE_METADATA_ID,
            resource_type="file",
            view_id=ViewId("cdf_cdm", "CogniteFile", "v1"),
            property_mapping=file_property_mapping,
        ),
        ResourceViewMappingApply(
            external_id=FILE_METADATA_EXTRACTOR_ID,
            resource_type="file",
            view_id=ViewId("cdf_extraction_extensions", "CogniteExtractorFile", "v1"),
            property_mapping={**file_property_mapping, "metadata": "extractedData"},
        ),
        ResourceViewMappingApply(
            external_id=ASSET_ANNOTATIONS_ID,
            resource_type="assetAnnotation",
            view_id=ViewId("cdf_cdm", "CogniteDiagramAnnotation", "v1"),
            property_mapping={
                # We are ignoring the symbol region in the default mapping.
                "annotatedResource.id": "edge.startNode",
                "annotationType": "edge.type.externalId",
                "creatingUser": "sourceCreatedUser",
                "creatingApp": "sourceId",
                "creatingAppVersion": "sourceContext",
                "status": "status",
                "data.assetRef.id": "edge.endNode",
                "data.assetRef.externalId": "edge.endNode",
                "data.description": "description",
                "data.pageNumber": "startNodePageNumber",
                "data.textRegion.confidence": "confidence",
                "data.textRegion.xMin": "startNodeXMin",
                "data.textRegion.xMax": "startNodeXMax",
                "data.textRegion.yMin": "startNodeYMin",
                "data.textRegion.yMax": "startNodeYMax",
                "data.text": "startNodeText",
            },
        ),
        ResourceViewMappingApply(
            external_id=FILE_ANNOTATIONS_ID,
            resource_type="fileAnnotation",
            view_id=ViewId("cdf_cdm", "CogniteFileAnnotation", "v1"),
            property_mapping={
                "annotatedResource.id": "edge.startNode",
                "annotationType": "edge.type.externalId",
                "creatingUser": "sourceCreatedUser",
                "creatingApp": "sourceId",
                "creatingAppVersion": "sourceContext",
                "status": "status",
                "data.fileRef.id": "edge.startNode",
                "data.fileRef.externalId": "edge.startNode",
                "data.description": "description",
                "data.pageNumber": "startNodePageNumber",
                "data.textRegion.confidence": "confidence",
                "data.textRegion.xMin": "startNodeXMin",
                "data.textRegion.xMax": "startNodeXMax",
                "data.textRegion.yMin": "startNodeYMin",
                "data.textRegion.yMax": "startNodeYMax",
                "data.text": "startNodeText",
            },
        ),
    ]



from cognite_toolkit._cdf_tk.client import ToolkitClient


class TestProjectAPI:
    def test_status(self, toolkit_client: ToolkitClient) -> None:
        result = toolkit_client.project.status()
        assert result.this_project.url_name == toolkit_client.config.project
        assert result.this_project.data_modeling_status in ("HYBRID", "DATA_MODELING_ONLY")



import sys
from types import MappingProxyType
from typing import Any, ClassVar, cast

from cognite.client.data_classes import WorkflowTrigger
from pydantic import Field, JsonValue, ModelWrapValidatorHandler, model_serializer, model_validator
from pydantic_core.core_schema import SerializationInfo, SerializerFunctionWrapHandler

from cognite_toolkit._cdf_tk.utils import humanize_collection

from .authentication import AuthenticationClientIdSecret
from .base import BaseModelResource, ToolkitResource

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


class TriggerRuleYAML(BaseModelResource):
    _trigger_type: ClassVar[str]

    @model_validator(mode="wrap")
    @classmethod
    def find_trigger_type(
        cls, data: "dict[str, Any] | TriggerRuleYAML", handler: ModelWrapValidatorHandler[Self]
    ) -> Self:
        if isinstance(data, TriggerRuleYAML):
            return cast(Self, data)
        if not isinstance(data, dict):
            raise ValueError(f"Invalid trigger rule data '{type(data)}' expected dict")

        if cls is not TriggerRuleYAML:
            # We are already in a subclass, so just validate as usual
            return handler(data)
        # If not we need to find the right subclass based on the triggerType field.
        if "triggerType" not in data:
            raise ValueError("Invalid trigger rule data missing 'triggerType' key")
        trigger_type = data["triggerType"]
        if trigger_type not in _TRIGGER_CLS_BY_NAME:
            raise ValueError(
                f"invalid trigger type '{trigger_type}'. Expected one of {humanize_collection(_TRIGGER_CLS_BY_NAME.keys(), bind_word='or')}"
            )
        cls_ = _TRIGGER_CLS_BY_NAME[trigger_type]
        return cast(Self, cls_.model_validate({k: v for k, v in data.items() if k != "triggerType"}))

    @model_serializer(mode="wrap", when_used="always", return_type=dict)
    def include_trigger_type(self, handler: SerializerFunctionWrapHandler) -> dict:
        if self._trigger_type is None:
            raise ValueError("Trigger type is not set")
        serialized_data = handler(self)
        serialized_data["triggerType"] = self._trigger_type
        return serialized_data


class ScheduleTrigger(TriggerRuleYAML):
    _trigger_type = "schedule"
    cron_expression: str = Field(
        description="A cron expression (UNIX format) specifying when the trigger should be executed. Use https://crontab.guru/ to create a cron expression. The API may adjust the exact timing of cron job executions to distribute the backend load more evenly. However, it will aim to maintain the overall frequency of executions as specified in the cron expression.",
    )
    timezone: str = Field(
        "UTC",
        description="Specifies the IANA time zone in which the cron expression is evaluated. Time zones must be valid as listed in https://docs.oracle.com/cd/E72987_01/wcs/tag-ref/MISC/TimeZones.html.",
    )


class DataModelingTrigger(TriggerRuleYAML):
    _trigger_type = "dataModeling"
    data_modeling_query: JsonValue
    batch_size: int = Field(
        ge=100, le=1_000, description="The maximum number of items to pass to a workflow execution."
    )
    batch_timeout: int = Field(
        ge=60,
        le=86_400,
        description="The maximum time in seconds to wait for the batch to be filled before passing it to a workflow execution.",
    )


class WorkflowTriggerYAML(ToolkitResource):
    _cdf_resource = WorkflowTrigger
    external_id: str = Field(
        max_length=255,
        description="Identifier for a trigger. Must be unique for the project. "
        "No trailing or leading whitespace and no null characters allowed.",
    )
    trigger_rule: TriggerRuleYAML
    input: JsonValue | None = None
    metadata: dict[str, str] | None = None
    workflow_external_id: str = Field(
        max_length=255,
        description="Identifier for a workflow. Must be unique for the project. "
        "No trailing or leading whitespace and no null characters allowed.",
    )
    workflow_version: str = Field(
        max_length=255,
        description="Identifier for a version. Must be unique for the workflow. No trailing or"
        " leading whitespace and no null characters allowed.",
    )
    authentication: AuthenticationClientIdSecret = Field(description="Credentials required for the authentication.")

    @model_serializer(mode="wrap")
    def serialize_trigger_rules(self, handler: SerializerFunctionWrapHandler, info: SerializationInfo) -> dict:
        # Trigger rules are serialized as empty dicts [{}, {}, ...]
        # This issue arises because Pydantic's serialization mechanism doesn't automatically
        # handle polymorphic serialization for subclasses of TriggerRuleYAML.
        # To address this, we include the below to explicitly calling model dump on the trigger rule
        serialized_data = handler(self)
        serialized_data["triggerRule"] = self.trigger_rule.model_dump(**vars(info))
        return serialized_data


_TRIGGER_CLS_BY_NAME: MappingProxyType[str, type[TriggerRuleYAML]] = MappingProxyType(
    {s._trigger_type: s for s in TriggerRuleYAML.__subclasses__()}
)



from abc import ABC
from typing import Any, Literal

from cognite.client import CogniteClient
from cognite.client.data_classes import FileMetadata, FileMetadataWrite, FileMetadataWriteList
from cognite.client.data_classes._base import (
    CogniteResourceList,
    ExternalIDTransformerMixin,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)


class _StreamlitCore(WriteableCogniteResource["StreamlitWrite"], ABC):
    def __init__(
        self,
        external_id: str,
        name: str,
        creator: str,
        entrypoint: str | None = None,
        description: str | None = None,
        published: bool = False,
        theme: Literal["Light", "Dark"] = "Light",
        thumbnail: str | None = None,
        data_set_id: int | None = None,
        cognite_toolkit_app_hash: str | None = None,
    ) -> None:
        self.external_id = external_id
        self.name = name
        self.creator = creator
        self.entrypoint = entrypoint
        self.description = description
        self.published = published
        self.theme = theme
        self.thumbnail = thumbnail
        self.data_set_id = data_set_id
        self.cognite_toolkit_app_hash = cognite_toolkit_app_hash

    def _as_file_args(self) -> dict[str, Any]:
        metadata = {
            "creator": self.creator,
            "description": self.description,
            "name": self.name,
            "published": self.published,
            "theme": self.theme,
        }
        if self.thumbnail:
            metadata["thumbnail"] = self.thumbnail
        if self.cognite_toolkit_app_hash:
            metadata["cdf-toolkit-app-hash"] = self.cognite_toolkit_app_hash
        if self.entrypoint:
            metadata["entrypoint"] = self.entrypoint
        return {
            "external_id": self.external_id,
            "name": f"{self.name}-source.json",
            "data_set_id": self.data_set_id,
            "directory": "/streamlit-apps/",
            "metadata": metadata,
        }


class StreamlitWrite(_StreamlitCore):
    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> "StreamlitWrite":
        args = dict(
            external_id=resource["externalId"],
            name=resource["name"],
            creator=resource["creator"],
            entrypoint=resource.get("entrypoint"),
            description=resource.get("description"),
            thumbnail=resource.get("thumbnail"),
            data_set_id=resource.get("dataSetId"),
        )
        # Trick to avoid specifying defaults twice
        for key in ["published", "theme"]:
            if key in resource:
                args[key] = resource[key]
        if "cogniteToolkitAppHash" in resource:
            args["cognite_toolkit_app_hash"] = resource["cogniteToolkitAppHash"]
        return cls(**args)

    def as_write(self) -> "StreamlitWrite":
        return self

    def as_file(self) -> FileMetadataWrite:
        return FileMetadataWrite(**self._as_file_args())


class Streamlit(_StreamlitCore):
    def __init__(
        self,
        external_id: str,
        name: str,
        creator: str,
        created_time: int,
        last_updated_time: int,
        entrypoint: str | None = None,
        description: str | None = None,
        published: bool = False,
        theme: Literal["Light", "Dark"] = "Light",
        thumbnail: str | None = None,
        data_set_id: int | None = None,
        cognite_toolkit_app_hash: str | None = None,
    ) -> None:
        super().__init__(
            external_id,
            name,
            creator,
            entrypoint,
            description,
            published,
            theme,
            thumbnail,
            data_set_id,
            cognite_toolkit_app_hash,
        )
        self.created_time = created_time
        self.last_updated_time = last_updated_time

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> "Streamlit":
        args = dict(
            external_id=resource["externalId"],
            name=resource["name"],
            creator=resource["creator"],
            entrypoint=resource.get("entrypoint"),
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            description=resource.get("description"),
            thumbnail=resource.get("thumbnail"),
            data_set_id=resource.get("dataSetId"),
        )
        # Trick to avoid specifying defaults twice
        for key in ["theme"]:
            if key in resource:
                args[key] = resource[key]
        if "cogniteToolkitAppHash" in resource:
            args["cognite_toolkit_app_hash"] = resource["cogniteToolkitAppHash"]
        if "published" in resource:
            if isinstance(resource["published"], str):
                args["published"] = resource["published"].strip().lower() == "true"
            else:
                args["published"] = resource["published"]
        return cls(**args)

    @classmethod
    def from_file(cls, file: FileMetadata) -> "Streamlit":
        dumped = file.dump()
        if "metadata" in dumped:
            dumped.update(dumped.pop("metadata"))
        if "cdf-toolkit-app-hash" in dumped:
            dumped["cogniteToolkitAppHash"] = dumped.pop("cdf-toolkit-app-hash")
        return cls._load(dumped)

    def as_write(self) -> StreamlitWrite:
        return StreamlitWrite(
            external_id=self.external_id,
            name=self.name,
            creator=self.creator,
            entrypoint=self.entrypoint,
            description=self.description,
            published=self.published,
            theme=self.theme,
            thumbnail=self.thumbnail,
            data_set_id=self.data_set_id,
            cognite_toolkit_app_hash=self.cognite_toolkit_app_hash,
        )

    def as_file(self) -> FileMetadata:
        args = self._as_file_args()
        args.update(
            created_time=self.created_time,
            last_updated_time=self.last_updated_time,
        )
        return FileMetadata(**args)


class StreamlitWriteList(CogniteResourceList[StreamlitWrite], ExternalIDTransformerMixin):
    _RESOURCE = StreamlitWrite

    def as_file_list(self) -> FileMetadataWriteList:
        return FileMetadataWriteList([item.as_file() for item in self])


class StreamlitList(WriteableCogniteResourceList[StreamlitWrite, Streamlit], ExternalIDTransformerMixin):
    _RESOURCE = Streamlit

    def as_write(self) -> StreamlitWriteList:
        return StreamlitWriteList([item.as_write() for item in self])

    def as_file_list(self) -> FileMetadataWriteList:
        return FileMetadataWriteList([item.as_file() for item in self])



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.streams import StreamYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic


def invalid_stream_test_cases() -> Iterable:
    yield pytest.param(
        {"settings": {"template": {"name": "ImmutableTestStream"}}},
        {"Missing required field: 'externalId'"},
        id="Missing required field: externalId",
    )
    yield pytest.param(
        {"externalId": "test-stream"},
        {"Missing required field: 'settings'"},
        id="Missing required field: settings",
    )
    yield pytest.param(
        {"externalId": "test-stream", "settings": {"template": {"name": "InvalidTemplate"}}},
        {
            "In settings.template.name input should be 'ImmutableTestStream', 'BasicArchive' or 'BasicLiveData'. Got 'InvalidTemplate'."
        },
        id="Invalid template name",
    )
    yield pytest.param(
        {
            "externalId": "test-stream",
            "settings": {"template": {"name": "ImmutableTestStream"}},
            "unknownField": "value",
        },
        {"Unused field: 'unknownField'"},
        id="Unused field",
    )
    yield pytest.param(
        {"externalId": "test-stream", "settings": {"template": {"names": "ImmutableTestStream"}}},
        {"In settings.template.names[key] input should be 'name'. Got 'names'."},
        id="wrong template dictionary format",
    )
    yield pytest.param(
        {"externalId": "Test-stream", "settings": {"template": {"name": "ImmutableTestStream"}}},
        {"In field externalId string should match pattern '^[a-z]([a-z0-9_-]{0,98}[a-z0-9])?$'"},
        id="Invalid external_id: starts with uppercase",
    )
    yield pytest.param(
        {"externalId": "test-stream-", "settings": {"template": {"name": "ImmutableTestStream"}}},
        {"In field externalId string should match pattern '^[a-z]([a-z0-9_-]{0,98}[a-z0-9])?$'"},
        id="Invalid external_id: ends with hyphen",
    )
    yield pytest.param(
        {"externalId": "test-stream_", "settings": {"template": {"name": "ImmutableTestStream"}}},
        {"In field externalId string should match pattern '^[a-z]([a-z0-9_-]{0,98}[a-z0-9])?$'"},
        id="Invalid external_id: ends with underscore",
    )
    yield pytest.param(
        {"externalId": "1test-stream", "settings": {"template": {"name": "ImmutableTestStream"}}},
        {"In field externalId string should match pattern '^[a-z]([a-z0-9_-]{0,98}[a-z0-9])?$'"},
        id="Invalid external_id: starts with number",
    )
    yield pytest.param(
        {"externalId": "t" * 101, "settings": {"template": {"name": "ImmutableTestStream"}}},
        {"In field externalId string should have at most 100 characters"},
        id="externalId-too-long",
    )


class TestStreamYAML:
    @pytest.mark.parametrize(
        "data",
        [
            pytest.param(
                {"externalId": "test-stream", "settings": {"template": {"name": "ImmutableTestStream"}}},
                id="ImmutableTestStream template",
            ),
            pytest.param(
                {"externalId": "archive-stream", "settings": {"template": {"name": "BasicArchive"}}},
                id="BasicArchive template",
            ),
            pytest.param(
                {"externalId": "live-stream", "settings": {"template": {"name": "BasicLiveData"}}},
                id="BasicLiveData template",
            ),
        ],
    )
    def test_valid_streams(self, data: dict) -> None:
        loaded = StreamYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_stream_test_cases()))
    def test_invalid_stream_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, StreamYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors
